# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
"""
Handle Thrift requests for authentication.
"""


import json
import codechecker_api_shared

from collections import defaultdict

from codechecker_api.Authentication_v6.ttypes import AccessControl, \
    AuthorisationList, HandshakeInformation, Permissions, SessionTokenData

from codechecker_common.logger import get_logger

from codechecker_server.profiler import timeit

from ..database.config_db_model import Product, ProductPermission, Session, \
    SystemPermission
from ..database.database import DBSession
from ..permissions import handler_from_scope_params as make_handler, \
    require_manager, require_permission
from ..server import permissions
from ..session_manager import generate_session_token

LOG = get_logger('server')


class ThriftAuthHandler:
    """
    Handle Thrift authentication requests.
    """

    def __init__(self, manager, auth_session, config_database):
        self.__manager = manager
        self.__auth_session = auth_session
        self.__config_db = config_database

    def __require_privilaged_access(self):
        """
        Checks if privilaged access is enabled for the server. Throws an
        exception if it is not.
        """
        if not self.getLoggedInUser():
            raise codechecker_api_shared.ttypes.RequestFailed(
                codechecker_api_shared.ttypes.ErrorCode.UNAUTHORIZED,
                "The server must be start by using privilaged access to "
                "execute this action.")

    def __has_permission(self, permission) -> bool:
        """ True if the current user has given permission rights. """
        if self.__manager.is_enabled and not self.__auth_session:
            return False

        return self.hasPermission(permission, None)

    def __require_permission_view(self):
        """
        Checks if the curret user has PERMISSION_VIEW rights. Throws an
        exception if it is not.
        """
        permission = codechecker_api_shared.ttypes.Permission.PERMISSION_VIEW
        if not self.__has_permission(permission):
            raise codechecker_api_shared.ttypes.RequestFailed(
                codechecker_api_shared.ttypes.ErrorCode.UNAUTHORIZED,
                "You are not authorized to execute this action.")

    @timeit
    def checkAPIVersion(self):
        # This is a deliberate empty call which if succeeds, marks for the
        # client that the server accepted the connection proper.
        pass

    # ============= Authentication and session handling =============

    @timeit
    def getAuthParameters(self):
        alive = self.__auth_session.is_alive if self.__auth_session \
                else False
        return HandshakeInformation(self.__manager.is_enabled, alive)

    @timeit
    def getLoggedInUser(self):
        if self.__auth_session:
            return self.__auth_session.user
        else:
            return ""

    @timeit
    def getAcceptedAuthMethods(self):
        return ["Username:Password"]

    @timeit
    def getAccessControl(self):
        self.__require_permission_view()

        with DBSession(self.__config_db) as session:
            global_permissions = Permissions(
                user=defaultdict(list),
                group=defaultdict(list))

            q = session.query(SystemPermission).all()
            for system_permission in q:
                name = system_permission.name
                perm = system_permission.permission
                if system_permission.is_group:
                    global_permissions.group[name].append(perm)
                else:
                    global_permissions.user[name].append(perm)

            product_permissions = {}
            q = session \
                .query(Product.endpoint, ProductPermission) \
                .outerjoin(Product,
                           ProductPermission.product_id == Product.id) \
                .all()

            for endpoint, product_permission in q:
                if endpoint not in product_permissions:
                    product_permissions[endpoint] = Permissions(
                        user=defaultdict(list),
                        group=defaultdict(list))

                name = product_permission.name
                perm = product_permission.permission
                if product_permission.is_group:
                    product_permissions[endpoint].group[name].append(perm)
                else:
                    product_permissions[endpoint].user[name].append(perm)

        default_superuser = self.__manager.default_superuser_name
        if default_superuser:
            global_permissions.user[default_superuser].append("SUPERUSER")

        return AccessControl(
            globalPermissions=global_permissions,
            productPermissions=product_permissions)

    @timeit
    def performLogin(self, auth_method, auth_string):
        if not auth_string:
            raise codechecker_api_shared.ttypes.RequestFailed(
                codechecker_api_shared.ttypes.ErrorCode.AUTH_DENIED,
                "No credentials supplied. Refusing authentication!")

        if auth_method == "Username:Password":
            user_name, _ = auth_string.split(':', 1)
            LOG.debug("'%s' logging in...", user_name)

            session = self.__manager.create_session(auth_string)

            if session:
                LOG.info("'%s' logged in.", user_name)
                return session.token
            else:
                msg = "Invalid credentials supplied for user '{0}'. " \
                      "Refusing authentication!".format(user_name)

                LOG.warning(msg)
                raise codechecker_api_shared.ttypes.RequestFailed(
                    codechecker_api_shared.ttypes.ErrorCode.AUTH_DENIED,
                    msg)

        raise codechecker_api_shared.ttypes.RequestFailed(
            codechecker_api_shared.ttypes.ErrorCode.AUTH_DENIED,
            "Could not negotiate via common authentication method.")

    @timeit
    def destroySession(self):
        user_name = self.getLoggedInUser()
        LOG.debug("'%s' logging out...", user_name)

        token = None
        if self.__auth_session:
            token = self.__auth_session.token

        is_logged_out = self.__manager.invalidate(token)
        if is_logged_out:
            LOG.info("'%s' logged out.", user_name)
        return is_logged_out

    # ============= Authorization, permission management =============

    @staticmethod
    def __unpack_extra_params(extra_params, session=None):
        """
        Helper function to unpack the extra_params JSON string to a dict.

        If specified, add the config_db_session to this dict too.
        """

        if extra_params and extra_params != "":
            params = json.loads(extra_params)
        else:
            params = {}

        if session:
            params['config_db_session'] = session

        return params

    @staticmethod
    def __create_permission_args(perm_enum, extra_params_string, session):
        """
        Helper function to transform the permission-specific values from the
        API into the appropriate Python constructs needed by the permission
        library.
        """

        perm = permissions.permission_from_api_enum(perm_enum)
        params = ThriftAuthHandler.__unpack_extra_params(extra_params_string,
                                                         session)
        return perm, params

    @timeit
    def getPermissions(self, scope):
        """
        Returns all the defined permissions in the given permission scope.
        """

        return [permissions.api_enum_for_permission(p)
                for p in permissions.get_permissions(scope)]

    @timeit
    def getPermissionsForUser(self, scope, extra_params, perm_filter):
        """
        Returns the permissions in the given permission scope and with the
        given scope-specific extra_params for the current logged in user,
        based on the permission filters.

        Filters in the perm_filter struct are joined in an AND clause.
        """

        if perm_filter is None or not any(perm_filter.__dict__.values()):
            # If no filtering is needed, this function behaves identically
            # to getPermissions().
            return self.getPermissions(scope)

        with DBSession(self.__config_db) as session:
            # The database connection must always be passed to the permission
            # handler.
            params = ThriftAuthHandler.__unpack_extra_params(extra_params,
                                                             session)

            perms = []
            for perm in permissions.get_permissions(scope):
                should_return = True
                handler = make_handler(perm, params)

                if should_return and perm_filter.given:
                    should_return = handler.has_permission(self.__auth_session)

                if should_return and perm_filter.canManage:
                    # If the user has any of the permissions that are
                    # authorised to manage the currently iterated permission,
                    # the filter passes.
                    should_return = require_manager(
                         perm, params, self.__auth_session)

                if should_return:
                    perms.append(perm)

            return [permissions.api_enum_for_permission(p)
                    for p in perms]

    @timeit
    def getAuthorisedNames(self, permission, extra_params):
        """
        Returns the users and groups who were EXPLICITLY granted a particular
        permission.
        """

        with DBSession(self.__config_db) as session:
            perm, params = ThriftAuthHandler.__create_permission_args(
                permission, extra_params, session)

            if not require_manager(perm, params, self.__auth_session):
                raise codechecker_api_shared.ttypes.RequestFailed(
                    codechecker_api_shared.ttypes.ErrorCode.UNAUTHORIZED,
                    "You can not manage the permission '{0}'"
                    .format(perm.name))

            handler = make_handler(perm, params)
            users, groups = handler.list_permitted()

            # The special default permission marker is an internal value.
            users = [user for user in users if user != '*']

            return AuthorisationList(users, groups)

    @timeit
    def addPermission(self, permission, auth_name, is_group, extra_params):
        """
        Adds the given permission to the user or group named auth_name.
        """

        with DBSession(self.__config_db) as session:
            perm, params = ThriftAuthHandler.__create_permission_args(
                permission, extra_params, session)

            if not require_manager(perm, params, self.__auth_session):
                raise codechecker_api_shared.ttypes.RequestFailed(
                    codechecker_api_shared.ttypes.ErrorCode.UNAUTHORIZED,
                    "You can not manage the permission '{0}'"
                    .format(perm.name))

            handler = make_handler(perm, params)
            handler.add_permission(auth_name.strip(),
                                   is_group,
                                   user_name=self.getLoggedInUser())

            session.commit()
            return True

    @timeit
    def removePermission(self, permission, auth_name, is_group, extra_params):
        """
        Removes the given permission from the user or group auth_name.
        """

        with DBSession(self.__config_db) as session:
            perm, params = ThriftAuthHandler.__create_permission_args(
                permission, extra_params, session)

            if not require_manager(perm, params, self.__auth_session):
                raise codechecker_api_shared.ttypes.RequestFailed(
                    codechecker_api_shared.ttypes.ErrorCode.UNAUTHORIZED,
                    "You can not manage the permission '{0}'"
                    .format(perm.name))

            handler = make_handler(perm, params)
            handler.remove_permission(auth_name, is_group,
                                      user_name=self.getLoggedInUser())

            session.commit()
            return True

    @timeit
    def hasPermission(self, permission, extra_params):
        """
        Returns whether or not the current logged-in user (or guest, if
        authentication is disabled) is granted the given permission.

        This method observes permission inheritance.
        """

        with DBSession(self.__config_db) as session:
            perm, params = ThriftAuthHandler.__create_permission_args(
                permission, extra_params, session)

            return require_permission(perm, params,
                                      self.__auth_session)

    # ============= Authorization, permission management =============

    @timeit
    def newToken(self, description):
        """
        Generate a new personal access token with the given description.
        """
        self.__require_privilaged_access()
        with DBSession(self.__config_db) as session:
            token = generate_session_token()
            user = self.getLoggedInUser()
            groups = ';'.join(self.__auth_session.groups)
            session_token = Session(token, user, groups, description, False)

            session.add(session_token)
            session.commit()

            LOG.info("New personal access token '%s...' has been generated "
                     "by '%s'.", token[:5], self.getLoggedInUser())

            return SessionTokenData(token,
                                    description,
                                    str(session_token.last_access))

    @timeit
    def removeToken(self, token):
        """
        Removes the given personal access token of the logged in user.
        """
        self.__require_privilaged_access()
        with DBSession(self.__config_db) as session:
            # Check if the given token is a personal access token so it can be
            # removed.
            user = self.getLoggedInUser()
            num_of_removed = session.query(Session) \
                .filter(Session.user_name == user) \
                .filter(Session.token == token) \
                .filter(Session.can_expire.is_(False)) \
                .delete(synchronize_session=False)
            session.commit()

            if not num_of_removed:
                raise codechecker_api_shared.ttypes.RequestFailed(
                    codechecker_api_shared.ttypes.ErrorCode.DATABASE,
                    "Personal access token {0} was not found in the "
                    "database.".format(token))

            # Invalidate the local session by token.
            self.__manager.invalidate_local_session(token)

            LOG.info("Personal access token '%s...' has been removed by '%s'.",
                     token[:5], self.getLoggedInUser())

            return True

    @timeit
    def getTokens(self):
        """
        Get available personal access tokens of the logged in user.
        """
        self.__require_privilaged_access()
        with DBSession(self.__config_db) as session:
            user = self.getLoggedInUser()
            sessionTokens = session.query(Session) \
                .filter(Session.user_name == user) \
                .filter(Session.can_expire.is_(False)) \
                .all()

            result = []
            for t in sessionTokens:
                result.append(SessionTokenData(
                    t.token,
                    t.description,
                    str(t.last_access)))

            return result



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
"""Parse the 'clang --version' command output."""

import os
import re
import shutil
import subprocess

from codechecker_analyzer import analyzer_context


class ClangVersionInfo:
    """ClangVersionInfo holds the version information of the used Clang."""

    def __init__(self,
                 major_version=None,
                 minor_version=None,
                 patch_version=None,
                 installed_dir=None,
                 vendor=None):

        self.major_version = int(major_version)
        self.minor_version = int(minor_version)
        self.patch_version = int(patch_version)
        self.installed_dir = str(installed_dir)
        self.vendor = str(vendor)


class ClangVersionInfoParser:
    """
    ClangVersionInfoParser is responsible for creating ClangVersionInfo
    instances from the version output of Clang.
    """

    def __init__(self, analyzer_binary='clang'):
        self.analyzer_binary = analyzer_binary
        self.clang_version_pattern = (
            r"(?P<vendor>clang|Apple LLVM) version (?P<major_version>[0-9]+)"
            r"\.(?P<minor_version>[0-9]+)\.(?P<patch_version>[0-9]+)")

        self.clang_installed_dir_pattern =\
            r"InstalledDir: (?P<installed_dir>[^\s]*)"

    def parse(self, version_string):
        """Try to parse the version string using the predefined patterns."""
        version_match = re.search(self.clang_version_pattern, version_string)
        if not version_match:
            return

        installed_dir_match = re.search(
            self.clang_installed_dir_pattern, version_string)

        if installed_dir_match:
            installed_dir = installed_dir_match.group('installed_dir')
        else:
            installed_dir = os.path.dirname(shutil.which(self.analyzer_binary))

        return ClangVersionInfo(
            version_match.group('major_version'),
            version_match.group('minor_version'),
            version_match.group('patch_version'),
            installed_dir,
            version_match.group('vendor'))


def get(clang_binary):
    """Get and parse the version information from given clang binary

    Should return False for getting the version
    information not from a clang compiler.
    """
    compiler_version = subprocess.check_output(
        [clang_binary, '--version'],
        env=analyzer_context.get_context().analyzer_env,
        encoding="utf-8",
        errors="ignore")
    version_parser = ClangVersionInfoParser(clang_binary)
    version_info = version_parser.parse(compiler_version)
    return version_info



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
"""
Argument handlers for the 'CodeChecker cmd product' subcommands.
"""


import sys

from codechecker_api.ProductManagement_v6.ttypes import DatabaseConnection, \
    ProductConfiguration

from codechecker_report_converter import twodim

from codechecker_common import logger

from codechecker_web.shared import database_status, convert

from .client import setup_product_client
from .cmd_line import CmdLineOutputEncoder
from .product import split_server_url

# Needs to be set in the handler functions.
LOG = None


def init_logger(level, stream=None, logger_name='system'):
    logger.setup_logger(level, stream)
    global LOG
    LOG = logger.get_logger(logger_name)


def handle_list_products(args):

    # If the given output format is not 'table', redirect logger's output to
    # the stderr.
    stream = None
    if 'output_format' in args and args.output_format != 'table':
        stream = 'stderr'

    init_logger(args.verbose if 'verbose' in args else None, stream)

    protocol, host, port = split_server_url(args.server_url)
    client = setup_product_client(protocol, host, port)
    products = client.getProducts(None, None)

    if args.output_format == 'json':
        results = []
        for product in products:
            results.append({product.endpoint: product})
        print(CmdLineOutputEncoder().encode(results))
    else:  # plaintext, csv
        header = ['Database status', 'Endpoint', 'Name', 'Description']
        rows = []
        for product in products:
            name = convert.from_b64(product.displayedName_b64) \
                if product.displayedName_b64 else ''
            description = convert.from_b64(product.description_b64) \
                if product.description_b64 else ''

            if not product.accessible:
                db_status_msg = 'No access.'
            else:
                db_status = product.databaseStatus
                db_status_msg = database_status.db_status_msg.get(
                    db_status, 'Unknown database status')

            rows.append((db_status_msg,
                         product.endpoint, name, description))

        print(twodim.to_str(args.output_format, header, rows))


def handle_add_product(args):

    init_logger(args.verbose if 'verbose' in args else None)

    protocol, host, port = split_server_url(args.server_url)
    client = setup_product_client(protocol, host, port)

    # Put together the database connection's descriptor.
    if 'postgresql' in args:
        db_engine = 'postgresql'
        db_host = args.dbaddress
        db_port = args.dbport
        db_user = args.dbusername
        db_pass = args.dbpassword
        db_name = args.dbname
    else:
        db_engine = 'sqlite'
        db_host = ""
        db_port = 0
        db_user = ""
        db_pass = ""
        db_name = args.sqlite

    dbc = DatabaseConnection(
        engine=db_engine,
        host=db_host,
        port=db_port,
        username_b64=convert.to_b64(db_user),
        password_b64=convert.to_b64(db_pass),
        database=db_name)

    # Put together the product configuration.
    name = convert.to_b64(args.display_name) \
        if 'display_name' in args else None
    desc = convert.to_b64(args.description) \
        if 'description' in args else None

    report_limit = None
    if hasattr(args, "report_limit") and args.report_limit:
        report_limit = int(args.report_limit)

    prod = ProductConfiguration(
        endpoint=args.endpoint,
        displayedName_b64=name,
        description_b64=desc,
        reportLimit=report_limit,
        connection=dbc)

    LOG.debug("Sending request to add product...")
    success = client.addProduct(prod)
    if success:
        LOG.info("Product added successfully.")
    else:
        LOG.error("Adding the product has failed.")
        sys.exit(1)


def handle_del_product(args):

    init_logger(args.verbose if 'verbose' in args else None)

    protocol, host, port = split_server_url(args.server_url)
    client = setup_product_client(protocol, host, port)

    # Endpoints substring-match.
    products = client.getProducts(args.endpoint, None)
    products = [product for product in products
                if product.endpoint == args.endpoint]

    if not products:
        LOG.error("The product '%s' does not exist!", args.endpoint)
        return

    success = client.removeProduct(products[0].id)
    if success:
        LOG.info("Product removed.")
    else:
        LOG.error("An error occurred in product removal.")
        sys.exit(1)



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
"""
Base class for various source analyzers.
"""


from abc import ABCMeta, abstractmethod
import os
import signal
import subprocess
import sys
import shlex

from typing import Optional

from codechecker_analyzer import analyzer_context
from codechecker_common.logger import get_logger

LOG = get_logger('analyzer')


class SourceAnalyzer(metaclass=ABCMeta):
    """
    Base class for different source analyzers.
    """

    def __init__(self, config_handler, buildaction):
        self.__config_handler = config_handler
        self.__build_action = buildaction
        # Currently analyzed source file.
        self.source_file = ''

    @property
    def buildaction(self):
        return self.__build_action

    @property
    def config_handler(self):
        return self.__config_handler

    @abstractmethod
    def construct_analyzer_cmd(self, result_handler):
        raise NotImplementedError("Subclasses should implement this!")

    @classmethod
    def resolve_missing_binary(cls, configured_binary, environ):
        """
        In case of the configured binary for the analyzer is not found in the
        PATH, this method is used to find a callable binary.
        """
        raise NotImplementedError("Subclasses should implement this!")

    @abstractmethod
    def get_binary_version(self, environ, details=False) -> str:
        """
        Return the version number of the binary that CodeChecker found, even
        if its incompatible. If details is true, additional version information
        is provided. If details is false, the return value should be
        convertible to a distutils.version.StrictVersion type.
        """
        raise NotImplementedError("Subclasses should implement this!")

    @classmethod
    def is_binary_version_incompatible(cls, environ) -> Optional[str]:
        """
        CodeChecker can only execute certain versions of analyzers.
        Returns a error object (an optional string). If the return value is
        None, the analyzer binary is compatible with the current CodeChecker
        version. Otherwise, the it should be a message describing the precise
        version mismatch.
        """
        raise NotImplementedError("Subclasses should implement this!")

    @classmethod
    def construct_config_handler(cls, args):
        """ Should return a subclass of AnalyzerConfigHandler."""
        raise NotImplementedError("Subclasses should implement this!")

    @abstractmethod
    def get_analyzer_mentioned_files(self, output):
        """
        Return a collection of files that were mentioned by the analyzer in
        its standard outputs, which should be analyzer_stdout or
        analyzer_stderr from a result handler.
        """
        raise NotImplementedError("Subclasses should implement this!")

    @abstractmethod
    def construct_result_handler(self, buildaction, report_output,
                                 skiplist_handler):
        """
        This method constructs the class that is responsible to handle the
        results of the analysis. The result should be a subclass of
        ResultHandler
        """
        raise NotImplementedError("Subclasses should implement this!")

    def analyze(self, analyzer_cmd, res_handler, proc_callback=None, env=None):
        """
        Run the analyzer.
        """
        LOG.debug('Running analyzer ...')

        LOG.debug_analyzer('\n%s',
                           ' '.join([shlex.quote(x) for x in analyzer_cmd]))

        res_handler.analyzer_cmd = analyzer_cmd
        try:
            ret_code, stdout, stderr \
                = SourceAnalyzer.run_proc(analyzer_cmd,
                                          res_handler.buildaction.directory,
                                          proc_callback,
                                          env)
            res_handler.analyzer_returncode = ret_code
            res_handler.analyzer_stdout = stdout
            res_handler.analyzer_stderr = stderr
            return res_handler

        except Exception as ex:
            LOG.error(ex)
            res_handler.analyzer_returncode = 1
            return res_handler

    @classmethod
    def get_analyzer_checkers(cls):
        """
        Return the checkers available in the analyzer.
        """
        raise NotImplementedError("Subclasses should implement this!")

    def post_analyze(self, result_handler):
        """
        Run immediately after the analyze function.
        """
        pass

    @staticmethod
    def run_proc(command, cwd=None, proc_callback=None, env=None):
        """
        Just run the given command and return the return code
        and the stdout and stderr outputs of the process.
        """

        def signal_handler(signum, frame):
            # Clang does not kill its child processes, so I have to.
            try:
                g_pid = proc.pid
                os.killpg(g_pid, signal.SIGTERM)
            finally:
                sys.exit(128 + signum)

        signal.signal(signal.SIGINT, signal_handler)

        if env is None:
            env = analyzer_context.get_context().analyzer_env

        proc = subprocess.Popen(
            command,
            bufsize=-1,
            env=env,
            preexec_fn=None if sys.platform == 'win32' else os.setsid,
            cwd=cwd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
            encoding="utf-8",
            errors="ignore")

        # Send the created analyzer process' object if somebody wanted it.
        if proc_callback:
            proc_callback(proc)

        stdout, stderr = proc.communicate()
        return proc.returncode, stdout, stderr



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
"""
Base Helper class for Thrift api calls.
"""

import sys
from thrift.transport import THttpClient
from thrift.protocol import TJSONProtocol

from codechecker_client.credential_manager import SESSION_COOKIE_NAME
from codechecker_client.product import create_product_url

from codechecker_common.logger import get_logger

LOG = get_logger('system')


class BaseClientHelper:

    def __init__(self, protocol, host, port, uri, session_token=None,
                 get_new_token=None):
        """
        @param get_new_token: a function which can generate a new token.
        """
        self.__host = host
        self.__port = port
        url = create_product_url(protocol, host, port, uri)

        self.transport = None

        try:
            self.transport = THttpClient.THttpClient(url)
        except ValueError:
            # Initalizing THttpClient may raise an exception if proxy settings
            # are used but the port number is not a valid integer.
            pass

        # Thrift do not handle the use case when invalid proxy format is
        # used (e.g.: no schema is specified). For this reason we need to
        # verify the proxy format in our side.
        self._validate_proxy_format()

        self.protocol = TJSONProtocol.TJSONProtocol(self.transport)
        self.client = None

        self.get_new_token = get_new_token
        self._set_token(session_token)

    def _validate_proxy_format(self):
        """
        Validate the proxy settings.
        If the proxy settings are invalid, it will print an error message and
        stop the program.
        """
        if self.transport and not self.transport.using_proxy():
            return

        if not self.transport or not self.transport.host or \
                not isinstance(self.transport.port, int):
            LOG.error("Invalid proxy format! Check your "
                      "HTTP_PROXY/HTTPS_PROXY environment variables if "
                      "these are in the right format: "
                      "'http[s]://host:port'.")
            sys.exit(1)

    def _set_token(self, session_token):
        """ Set the given token in the transport layer. """
        if not session_token:
            return

        headers = {'Cookie': SESSION_COOKIE_NAME + '=' + session_token}
        self.transport.setCustomHeaders(headers)

    def _reset_token(self):
        """ Get a new token and update the transport layer. """
        if not self.get_new_token:
            return

        # get_new_token() function connects to a remote server to get a new
        # session token.
        session_token = self.get_new_token()
        self._set_token(session_token)



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------

import logging
import os
import re

from copy import deepcopy
from typing import Iterator, List, Tuple

from codechecker_report_converter.report import BugPathEvent, \
    get_or_create_file, Report
from ..parser import BaseParser


LOG = logging.getLogger('report-converter')


class Parser(BaseParser):
    """ Parser for clang-tidy console output. """

    def __init__(self):
        super(Parser, self).__init__()

        # Regex for parsing a clang-tidy message.
        self.message_line_re = re.compile(
            # File path followed by a ':'.
            r'^(?P<path>[\S ]+?):'
            # Line number followed by a ':'.
            r'(?P<line>\d+?):'
            # Column number followed by a ':' and a space.
            r'(?P<column>\d+?): '
            # Severity followed by a ':'.
            r'(?P<severity>(error|warning)):'
            # Checker message.
            r'(?P<message>[\S \t]+)\s*'
            # Checker name.
            r'\[(?P<checker>.*)\]')

        # Matches a note.
        self.note_line_re = re.compile(
            # File path followed by a ':'.
            r'^(?P<path>[\S ]+?):'
            # Line number followed by a ':'.
            r'(?P<line>\d+?):'
            # Column number followed by a ':' and a space.
            r'(?P<column>\d+?): '
            # Severity == note.
            r'note:'
            # Checker message.
            r'(?P<message>.*)')

        # Matches pre Clang 17 fix-its:
        # "       fixit-text"
        self.fixit_old_re = re.compile(
            r'^\s+(?P<message>\S.*)')

        # Matches post clang 17 fix-its
        # "   28 |     fixit-text"
        self.fixit_new_re = re.compile(
            r'^\s*\d*\s*\|\s*(?P<message>\S.*)')

    def _parse_line(
        self,
        it: Iterator[str],
        line: str
    ) -> Tuple[List[Report], str]:
        """ Parse the given line. """
        match = self.message_line_re.match(line)
        if match is None:
            return [], next(it)

        checker_names = match.group('checker').strip().split(",")
        report = Report(
            get_or_create_file(
                os.path.abspath(match.group('path')), self._file_cache),
            int(match.group('line')),
            int(match.group('column')),
            match.group('message').strip(),
            checker_names[0],
            bug_path_events=[])

        try:
            line = next(it)
            line = self._parse_code(it, line)
            line = self._parse_fixits(report, it, line)
            line = self._parse_notes(report, it, line)
        except StopIteration:
            line = ''
        finally:
            report.bug_path_events.append(BugPathEvent(
                report.message,
                report.file,
                report.line,
                report.column))

            # When a checker name and the alias of this checker is turned on,
            # Clang Tidy (>v11) will generate only one report where the checker
            # names are concatenated with ',' mark. With this we will generate
            # multiple reports for each checker name / alias.
            reports = []
            for checker_name in checker_names:
                r = deepcopy(report)
                r.checker_name = checker_name
                r.category = self._get_category(checker_name)

                reports.append(r)

            return reports, line

    def _get_category(self, checker_name: str) -> str:
        """ Get category for Clang-Tidy checker. """
        parts = checker_name.split('-')
        return parts[0] if parts else 'unknown'

    def _parse_code(
        self,
        it: Iterator[str],
        line: str
    ) -> str:
        # Eat code line.
        if self.note_line_re.match(line) or self.message_line_re.match(line):
            LOG.debug("Unexpected line: %s. Expected a code line!", line)
            return line

        # Eat arrow line.
        # FIXME: range support?
        line = next(it)
        if '^' not in line:
            LOG.debug("Unexpected line: %s. Expected an arrow line!", line)
            return line

        return next(it)

    def _parse_fixits(
        self,
        report: Report,
        it: Iterator[str],
        line: str
    ) -> str:
        """ Parses fixit messages. """

        while self.message_line_re.match(line) is None and \
                self.note_line_re.match(line) is None:
            match = self.fixit_new_re.match(line)
            if not match:
                match = self.fixit_old_re.match(line)
                message_text = match.group("message")
                # Until Clang 16, the FixIt line starts with whitespace.
                col = line.find(message_text) + 1
            else:
                message_text = match.group("message")
                # In newer versions, we have whitespace then, optionally
                # a line number, and then a | character.
                col = line.find(message_text) - line.find("|") - 1

            report.notes.append(BugPathEvent(
                    f"{message_text} (fixit)",
                    report.file,
                    report.line,
                    col))
            line = next(it)
        return line

    def _parse_notes(
        self,
        report: Report,
        it: Iterator[str],
        line: str
    ) -> str:
        """ Parses note messages. """

        while self.message_line_re.match(line) is None:
            match = self.note_line_re.match(line)
            if match is None:
                LOG.debug("Unexpected line: %s", line)
                return next(it)

            report.bug_path_events.append(BugPathEvent(
                match.group('message').strip(),
                get_or_create_file(
                    os.path.abspath(match.group('path')), self._file_cache),
                int(match.group('line')),
                int(match.group('column'))))

            line = next(it)
            line = self._parse_code(it, line)
        return line



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
"""Codeclimate output helpers."""

from typing import Dict, List

from codechecker_report_converter.report import Report


def convert(reports: List[Report]) -> List[Dict]:
    """Convert the given reports to codeclimate format.

    This function will convert the given report to Code Climate format.
    reports - list of reports type Report

    returns a list of reports converted to codeclimate format
    """
    codeclimate_reports = []

    for report in reports:
        codeclimate_reports.append(__to_codeclimate(report))

    return codeclimate_reports


__codeclimate_severity_map = {
    'CRITICAL': 'critical',
    'HIGH': 'major',
    'MEDIUM': 'minor',
    'LOW': 'minor',
    'STYLE': 'info',
    'UNSPECIFIED': 'info',
}


def __to_codeclimate(report: Report) -> Dict:
    """Convert a Report to Code Climate format."""
    severity = __codeclimate_severity_map.get(report.severity) \
        if report.severity else 'info'

    return {
        "type": "issue",
        "check_name": report.checker_name,
        "description": report.message,
        "categories": ["Bug Risk"],
        "fingerprint": report.report_hash,
        "severity": severity,
        "location": {
            "path": report.file.path,
            "lines": {
                "begin": report.line
            }
        }
    }



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------

"""
Main CodeChecker script.
"""


import argparse
from importlib import machinery
import json
import os
import signal
import sys


class ArgumentParser(argparse.ArgumentParser):
    def error(self, message):
        """
        By default argparse will exit with error code 2 in case of error, but
        we are using this error code for different purposes. For this reason
        we will override this function, so we will exit with error code 1 in
        such cases too.
        """
        self.print_usage(sys.stderr)
        self.exit(1, f"{self.prog}: error: {message}\n")


def add_subcommand(subparsers, sub_cmd, cmd_module_path, lib_dir_path):
    """
    Load the subcommand module and then add the subcommand to the available
    subcommands in the given subparsers collection.

    subparsers has to be the return value of the add_parsers() method on an
    argparse.ArgumentParser.
    """
    m_path, m_name = os.path.split(cmd_module_path)

    module_name = os.path.splitext(m_name)[0]
    target = [os.path.join(lib_dir_path, m_path)]

    # Load the module named as the argument.
    cmd_spec = machinery.PathFinder().find_spec(module_name,
                                                target)
    command_module = cmd_spec.loader.load_module(module_name)

    # Now that the module is loaded, construct an ArgumentParser for it.
    sc_parser = subparsers.add_parser(
        sub_cmd, **command_module.get_argparser_ctor_args())

    # Run the method which adds the arguments to the subcommand's handler.
    command_module.add_arguments_to_parser(sc_parser)


def get_data_files_dir_path():
    """ Get data files directory path """
    bin_dir = os.environ.get('CC_BIN_DIR')

    # In case of internal package we return the parent directory of the 'bin'
    # folder.
    if bin_dir:
        return os.path.dirname(bin_dir)

    # If this is a pip-installed package, try to find the data directory.
    import sysconfig
    data_dir_paths = [
        # Try to find the data directory beside the lib directory.
        # /usr/local/lib/python3.8/dist-packages/codechecker_common/cli.py
        # (this file) -> /usr/local
        os.path.abspath(os.path.join(__file__, *[os.path.pardir] * 5)),

        # Automatically try to find data directory if it can be found in
        # standard locations.
        *set(sysconfig.get_path("data", s)
             for s in sysconfig.get_scheme_names())]

    for dir_path in data_dir_paths:
        data_dir_path = os.path.join(dir_path, 'share', 'codechecker')
        if os.path.exists(data_dir_path):
            return data_dir_path

    print("Failed to get CodeChecker data files directory path in: ",
          data_dir_paths)
    sys.exit(1)


def main():
    """
    CodeChecker main command line.
    """
    if not os.environ.get('CC_LIB_DIR'):
        os.environ['CC_LIB_DIR'] = \
                os.path.dirname(os.path.dirname(os.path.realpath(__file__)))

    data_files_dir_path = get_data_files_dir_path()
    os.environ['CC_DATA_FILES_DIR'] = data_files_dir_path

    # Load the available CodeChecker subcommands.
    # This list is generated dynamically by scripts/build_package.py, and is
    # always meant to be available alongside the CodeChecker.py.
    commands_cfg = os.path.join(data_files_dir_path, "config", "commands.json")

    with open(commands_cfg, encoding="utf-8", errors="ignore") as cfg_file:
        subcommands = json.load(cfg_file)

    def signal_handler(signum, frame):
        """
        Without this handler the PostgreSQL
        server does not terminate at signal.
        """
        sys.exit(128 + signum)

    signal.signal(signal.SIGINT, signal_handler)
    signal.signal(signal.SIGTERM, signal_handler)

    try:
        parser = ArgumentParser(
            prog="CodeChecker",
            formatter_class=argparse.RawDescriptionHelpFormatter,
            description="""Run the CodeChecker sourcecode analyzer framework.
Please specify a subcommand to access individual features.""",
            epilog="""Example scenario: Analyzing, and storing results
------------------------------------------------
Start the server where the results will be stored and can be viewed
after the analysis is done:
    CodeChecker server

Analyze a project with default settings:
    CodeChecker check -b "cd ~/myproject && make" -o "~/results"

Store the analyzer results to the server:
    CodeChecker store "~/results" -n myproject

The results can be viewed:
 * In a web browser: http://localhost:8001
 * In the command line:
    CodeChecker cmd results myproject

Example scenario: Analyzing, and printing results to Terminal (no storage)
--------------------------------------------------------------------------
In this case, no database is used, and the results are printed on the standard
output.

    CodeChecker check -b "cd ~/myproject && make\" """)

        subparsers = parser.add_subparsers(help='commands')

        if subcommands:
            # Try to check if the user has already given us a subcommand to
            # execute. If so, don't load every available parts of CodeChecker
            # to ensure a more optimised run.
            if len(sys.argv) > 1:
                first_command = sys.argv[1]
                if first_command in subcommands:

                    # Consider only the given command as an available one.
                    subcommands = {first_command: subcommands[first_command]}

            lib_dir_path = os.environ.get('CC_LIB_DIR')
            for subcommand in subcommands:
                try:
                    add_subcommand(subparsers, subcommand,
                                   subcommands[subcommand], lib_dir_path)
                except (IOError, ImportError):
                    print("Couldn't import module for subcommand '" +
                          subcommand + "'... ignoring.")
                    import traceback
                    traceback.print_exc(file=sys.stdout)

        args = parser.parse_args()

        # Call handler function to process configuration files. If there are
        # any configuration options available in one of the given file than
        # extend the system argument list with these options and try to parse
        # the argument list again to validate it.
        if 'func_process_config_file' in args:
            # Import logger module here after 'CC_DATA_FILES_DIR' environment
            # variable is set, so 'setup_logger' will be able to initialize
            # the logger properly.
            from codechecker_common import logger
            logger.setup_logger(
                args.verbose if 'verbose' in args else None,
                'stderr')
            LOG = logger.get_logger('system')

            if len(sys.argv) > 1:
                called_sub_command = sys.argv[1]

            cfg_args = args.func_process_config_file(args, called_sub_command)
            if cfg_args:
                # Expand environment variables in the arguments.
                cfg_args = [os.path.expandvars(cfg) for cfg in cfg_args]

                # Replace --config option with the options inside the config
                # file.
                cfg_idx = sys.argv.index("--config")
                sys.argv = sys.argv[:cfg_idx] + cfg_args + \
                    sys.argv[cfg_idx + 2:]

                args = parser.parse_args()
                LOG.info("Full extended command: %s", ' '.join(sys.argv))

        if 'func' in args:
            sys.exit(args.func(args))
        else:
            # Print the help message of the current command if no subcommand
            # is given.
            sys.argv.append("--help")
            args = parser.parse_args()
            args.func(args)

    except KeyboardInterrupt as kb_err:
        print(str(kb_err))
        print("Interrupted by user...")
        sys.exit(1)

    # Handle all exception, but print stacktrace. It is needed for atexit.
    # atexit does not work correctly when an unhandled exception occurred.
    # So in this case, the servers left running when the script exited.
    except Exception:
        import traceback
        traceback.print_exc(file=sys.stdout)
        sys.exit(1)


# -----------------------------------------------------------------------------
if __name__ == "__main__":
    main()



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------

""" Special return value statistics collector. """

from io import StringIO

from collections import defaultdict
import os
import re


class SpecialReturnValueCollector:
    """ Collect special return value statistics.

    This script lists functions of which the return value is checked for
    negative (integers) or null (pointers).
    """

    # Checker name used for pre analysis.
    checker_collect = 'statisticscollector.SpecialReturnValue'

    # Checker name which runs the analysis.
    checker_analyze = 'statisticsbased.SpecialReturnValue'

    def __init__(self, stats_min_sample_count, stats_relevance_threshold):
        self.stats_min_sample_count = stats_min_sample_count
        self.stats_relevance_threshold = stats_relevance_threshold

        # Matching these lines:
        # /.../x.c:551:12: warning:
        # Special Return Value:/.../x.c:551:12,parsedate,0,0

        ptrn = \
            r'.*warning: Special Return Value:'\
            '.*:[0-9]*:[0-9]*.*,(.*),([0,1]),([0,1])'
        self.special_ret_val_regexp = re.compile(ptrn)

        self.stats = {'total': defaultdict(int),
                      'nof_negative': defaultdict(int),
                      'nof_null': defaultdict(int)}

    @staticmethod
    def stats_file(path):
        return os.path.join(path, 'SpecialReturn.yaml')

    @staticmethod
    def checker_analyze_cfg(path):
        """ Return the checker config parameter for the analyzer checker. """
        return ['-Xclang', '-analyzer-config',
                '-Xclang',
                'alpha.ericsson.statisticsbased:APIMetadataPath=' + path]

    def total(self):
        return self.stats.get('total')

    def nof_null(self):
        return self.stats.get('nof_null')

    def nof_negative(self):
        return self.stats.get('nof_negative')

    def process_line(self, line):
        """ Match regex on the line. """
        m = self.special_ret_val_regexp.match(line)
        if m:
            func = m.group(1)
            ret_negative = m.group(2)
            ret_null = m.group(3)

            self.stats['total'][func] += 1
            self.stats['nof_negative'][func] += int(ret_negative)
            self.stats['nof_null'][func] += int(ret_null)

    def filter_stats(self):
        """ Filter the collected statistics based on the threshold. """
        neg = []
        null = []
        stats = self.stats
        total = stats.get('total')

        for key in sorted(stats.get('total').keys()):
            negative_ratio = stats['nof_negative'][key]/stats['total'][key]
            if (self.stats_relevance_threshold < negative_ratio < 1 and
                    total[key] >= self.stats_min_sample_count):
                neg.append(key)

            null_ratio = stats['nof_null'][key]/stats['total'][key]
            if (self.stats_relevance_threshold < null_ratio < 1 and
                    total[key] >= self.stats_min_sample_count):
                null.append(key)
        return neg, null

    def get_yaml(self):
        """ Get statistics in yaml format.

        FIXME proper yaml generation.
        """
        stats_yaml = StringIO()

        stats_yaml.write("#\n")
        stats_yaml.write("# SpecialReturn metadata format 1.0\n")
        neg, null = self.filter_stats()

        for n in neg:
            stats_yaml.write(
                "{name: " + n + ", relation: LT, value: 0}\n")
        for n in null:
            stats_yaml.write(
                "{name: " + n + ", relation: EQ, value: 0}\n")

        return stats_yaml.getvalue()



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
from collections import defaultdict
import os
from typing import Any, cast, DefaultDict, Dict, Iterable, List, Optional, \
    Set, Tuple, Union

from codechecker_common.util import load_json


def split_label_kv(key_value: str) -> Tuple[str, str]:
    """
    A label has a value separated by colon (:) character, e.g:
    "severity:high". This function returns this key and value as a tuple.
    Optional whitespaces around (:) and at the two ends of this string are
    not taken into account. If key_value contains no colon, then the value
    is empty string.
    """
    try:
        pos = key_value.index(':')
    except ValueError:
        return (key_value.strip(), '')

    return key_value[:pos].strip(), key_value[pos + 1:].strip()


# TODO: Most of the methods of this class get an optional analyzer name. If
# None is given to these functions then labels of any analyzer's checkers is
# taken into account. This the union of all analyzers' checkers is a bad
# approach because different analyzers may theoretically have checkers with the
# same name. Fortunately this is not the case with the current analyzers'
# checkers, so now it works properly.
class CheckerLabels:
    # These labels should be unique by each analyzer. If this uniqueness is
    # violated then an error is thrown during label JSON file parse. The
    # default value of the label also needs to be provided here.
    UNIQUE_LABELS = {
        'severity': 'UNSPECIFIED'}

    def __init__(self, checker_labels_dir: str):
        if not os.path.isdir(checker_labels_dir):
            raise NotADirectoryError(
                f'{checker_labels_dir} is not a directory.')

        label_json_files: Iterable[str] = os.listdir(
            os.path.join(checker_labels_dir, 'analyzers'))

        self.__descriptions = {}

        if 'descriptions.json' in os.listdir(checker_labels_dir):
            self.__descriptions = load_json(os.path.join(
                checker_labels_dir, 'descriptions.json'))

        label_json_files = map(
            lambda f: os.path.join(checker_labels_dir, 'analyzers', f),
            label_json_files)

        self.__data = self.__union_label_files(label_json_files)
        self.__check_json_format(self.__data)

    def __union_label_files(
        self,
        label_files: Iterable[str]
    ) -> Dict[str, DefaultDict[str, List[str]]]:
        """
        This function creates a union object of the given label files. The
        resulting object maps analyzers to the collection of their checkers
        with their labels:

        {
            "analyzer1": {
                "checker1": [ ... ]
                "checker2": [ ... ]
            },
            "analyzer2": {
                ...
            }
        }
        """
        all_labels = {}

        for label_file in label_files:
            data = load_json(label_file)
            analyzer_labels = defaultdict(list)

            for checker, labels in data['labels'].items():
                analyzer_labels[checker].extend(labels)

            all_labels[data['analyzer']] = analyzer_labels

        return all_labels

    def __check_json_format(self, data: dict):
        """
        Check the format of checker labels' JSON config file, i.e. this file
        must contain specific values with specific types. For example the
        checker labels are string lists. In case of any format error a
        ValueError exception is thrown with the description of the wrong
        format.
        """
        def is_string_list(x):
            return isinstance(x, list) and \
                all(map(lambda s: isinstance(s, str), x))

        def is_unique(labels: Iterable[str], label: str):
            """
            Check if the given label occurs only once in the label list.
            """
            found = False
            for k, _ in map(split_label_kv, labels):
                if k == label:
                    if found:
                        return False
                    else:
                        found = True
            return True

        if not isinstance(data, dict):
            raise ValueError(f'Top level element should be a JSON object.')

        for _, checkers in data.items():
            for checker, labels in checkers.items():
                if not is_string_list(labels):
                    raise ValueError(
                        f'"{checker}" should be assigned a string list.')

                if any(map(lambda s: ':' not in s, labels)):
                    raise ValueError(
                        f'All labels at "{checker}" should be in the '
                        'following format: <label>:<property>.')

                for unique_label in CheckerLabels.UNIQUE_LABELS:
                    if not is_unique(labels, unique_label):
                        raise ValueError(
                            f'Label "severity" should be unique for checker '
                            '{checker}.')

    def __get_analyzer_data(
        self,
        analyzer: Optional[str] = None
    ) -> Iterable[Tuple[str, Any]]:
        """
        Most functions of this class require an analyzer name which determines
        a checker name specifically. If no analyzer is given then all of them
        is taken into account (for backward-compatibility reasons). This helper
        function yields either an analyzer's label data or all analyzer's
        label data.
        """
        for a, c in self.__data.items():
            if analyzer is None or a == analyzer:
                yield a, c

    def get_analyzers(self) -> Iterable[str]:
        return self.__data.keys()

    def checkers_by_labels(
        self,
        filter_labels: Iterable[str],
        analyzer: Optional[str] = None
    ) -> List[str]:
        """
        Returns a list of checkers that have at least one of the specified
        labels.

        filter_labels -- A string list which contains labels with specified
                         values. E.g. ['profile:default', 'severity:high'].
        analyzer -- An optional analyzer name of which checkers are searched.
                    By default all analyzers are searched.
        """
        collection = []

        label_set = set(map(split_label_kv, filter_labels))

        for _, checkers in self.__get_analyzer_data(analyzer):
            for checker, labels in checkers.items():
                labels = set(map(split_label_kv, labels))

                if labels.intersection(label_set):
                    collection.append(checker)

        return collection

    def label_of_checker(
        self,
        checker: str,
        label: str,
        analyzer: Optional[str] = None
    ) -> Union[str, List[str]]:
        """
        If a label has unique constraint then this function retuns the value
        that belongs to the given label or the default value that is set among
        label constraints. If there is no unique constraint then an iterable
        object returns with the values assigned to the given label. If the
        checker name is not found in the label config file then its prefixes
        are also searched. For example "clang-diagnostic" in the config file
        matches "clang-diagnostic-unused-argument".
        """
        labels = (
            value
            for key, value in self.labels_of_checker(checker, analyzer)
            if key == label)

        if label in CheckerLabels.UNIQUE_LABELS:
            try:
                return next(labels)
            except StopIteration:
                return CheckerLabels.UNIQUE_LABELS[label]

        # TODO set() is used for uniqueing results in case a checker name is
        # provided by multiple analyzers. This will be unnecessary when we
        # cover this case properly.
        return list(set(labels))

    def severity(self, checker: str, analyzer: Optional[str] = None) -> str:
        """
        Shorthand for the following call:
        checker_labels.label_of_checker(checker, 'severity', analyzer)
        """
        return cast(str, self.label_of_checker(checker, 'severity', analyzer))

    def labels_of_checker(
        self,
        checker: str,
        analyzer: Optional[str] = None
    ) -> List[Tuple[str, str]]:
        """
        Return the list of labels of a checker. The list contains (label,
        value) pairs. If the checker name is not found in the label config file
        then its prefixes are also searched. For example "clang-diagnostic" in
        the config file matches "clang-diagnostic-unused-argument".
        """
        labels: List[Tuple[str, str]] = []

        for _, checkers in self.__get_analyzer_data(analyzer):
            c: Optional[str] = checker

            if c not in checkers:
                c = next(filter(
                    lambda c: checker.startswith(cast(str, c)),
                    iter(checkers.keys())), None)

            labels.extend(map(split_label_kv, checkers.get(c, [])))

        # TODO set() is used for uniqueing results in case a checker name is
        # provided by multiple analyzers. This will be unnecessary when we
        # cover this case properly.
        return list(set(labels))

    def get_description(self, label: str) -> Optional[Dict[str, str]]:
        """
        Returns the descriptions of the given label's values.
        """
        return self.__descriptions.get(label)

    def checkers(self, analyzer: Optional[str] = None) -> List[str]:
        """
        Return the list of available checkers.
        """
        collection = []

        for _, checkers in self.__get_analyzer_data(analyzer):
            collection.extend(checkers.keys())

        return collection

    def labels(self, analyzer: Optional[str] = None) -> List[str]:
        """
        Returns a list of occurring labels.
        """
        collection: Set[str] = set()

        for _, checkers in self.__get_analyzer_data(analyzer):
            for labels in checkers.values():
                collection.update(map(
                    lambda x: split_label_kv(x)[0], labels))

        return list(collection)

    def occurring_values(
        self,
        label: str,
        analyzer: Optional[str] = None
    ) -> List[str]:
        """
        Return the list of values belonging to the given label which were used
        for at least one checker.
        """
        values = set()

        for _, checkers in self.__get_analyzer_data(analyzer):
            for labels in checkers.values():
                for lab, value in map(split_label_kv, labels):
                    if lab == label:
                        values.add(value)

        return list(values)



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
"""
Database connection handling to a database backend.
"""
from abc import ABCMeta, abstractmethod
import os
import subprocess

from alembic import command, config
from alembic import script
from alembic import migration

from alembic.util import CommandError
import sqlalchemy
from sqlalchemy import event
from sqlalchemy.engine.url import URL, make_url
from sqlalchemy.orm import sessionmaker
from sqlalchemy.pool import NullPool

from codechecker_api_shared.ttypes import DBStatus

from codechecker_common.logger import get_logger

from codechecker_web.shared import host_check, pgpass


LOG = get_logger('system')


def call_command(cmd, env=None, cwd=None):
    """ Call an external cmd and return with (output, return_code)."""

    try:
        LOG.debug('Run %s', ' '.join(cmd))
        out = subprocess.check_output(cmd,
                                      env=env,
                                      stderr=subprocess.STDOUT,
                                      cwd=cwd,
                                      encoding="utf-8",
                                      errors="ignore")
        LOG.debug(out)
        return out, 0
    except subprocess.CalledProcessError as ex:
        LOG.debug('Running command "%s" Failed.', ' '.join(cmd))
        LOG.debug(str(ex.returncode))
        LOG.debug(ex.output)
        return ex.output, ex.returncode
    except OSError as oerr:
        LOG.warning(oerr.strerror)
        return oerr.strerror, oerr.errno


class DBSession:
    """
    Requires a session maker object and creates one session which can be used
    in the context.

    The session will be automatically closed, but commiting must be done
    inside the context.
    """
    def __init__(self, session_maker):
        self.__session = None
        self.__session_maker = session_maker

    def __enter__(self):
        # create new session
        self.__session = self.__session_maker()
        return self.__session

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.__session:
            self.__session.close()


class DBContext:
    """
    Simple helper class to setup and sql engine, a database session
    and a connection.

    NOTE: The error property should be always checked inside the
    with statement to verify that the engine/connection/session
    setup was successful.
    """

    def __init__(self, sql_server):
        self.sql_server = sql_server
        self.db_engine = None
        self.db_session = None
        self.db_connection = None
        self.db_error = None

    def __enter__(self):
        self.db_error = None
        try:
            self.db_engine = self.sql_server.create_engine()
            self.db_session = sessionmaker(bind=self.engine)()
            self.db_connection = self.db_session.connection()
        except Exception as ex:
            LOG.debug("Connection error:")
            LOG.debug("%s", str(ex))
            import traceback
            traceback.print_exc()
            self.db_error = DBStatus.FAILED_TO_CONNECT

        return self

    @property
    def error(self):
        # Indicate that there is some problem with the
        # database setup.
        return self.db_error

    @property
    def engine(self):
        # Get the configured engine.
        return self.db_engine

    @property
    def connection(self):
        # Get the configured connection.
        return self.db_connection

    @property
    def session(self):
        # Get the configured session.
        return self.db_session

    def __exit__(self, *args):
        if self.db_session:
            self.db_session.close()
        if self.db_connection:
            self.db_connection.close()
        if self.db_engine:
            self.db_engine.dispose()


class SQLServer(metaclass=ABCMeta):
    """
    Abstract base class for database server handling. An SQLServer instance is
    responsible for the connection management towards the database.

    SQLServer implementations are created via SQLServer.from_cmdline_args().

    How to add a new database server implementation:
        1, Derive from SQLServer and implement the abstract methods
        2, Add/modify some command line options in CodeChecker.py
        3, Modify SQLServer.from_cmdline_args() in order to create an
           instance of the new server type if needed
    """

    def __init__(self, name_in_log: str, model_meta, migration_root):
        """
        Sets self.migration_root. migration_root should be the path to the
        alembic migration scripts.

        Also sets the created class' model identifier to the given meta dict.
        """
        self.name_in_log = name_in_log
        self.__model_meta = model_meta
        self.migration_root = migration_root

    def _create_schema(self):
        """
        Creates the database schema if needed.
        The database server has to be started before this method is called.
        """

        try:
            with DBContext(self) as db:
                if db.error:
                    return db.error

                cfg = config.Config()
                cfg.set_main_option("script_location", self.migration_root)
                cfg.attributes["connection"] = db.connection
                cfg.attributes["database_name"] = self.name_in_log

                mcontext = migration.MigrationContext.configure(db.connection)
                database_schema_revision = mcontext.get_current_revision()
                LOG.debug('Schema revision in the database: ' +
                          str(database_schema_revision))

                if database_schema_revision:
                    LOG.debug('Database schema was found.'
                              ' No need to initialize new.')
                else:
                    LOG.debug('No schema was detected in the database.')
                    LOG.debug('Initializing new ...')
                    command.upgrade(cfg, "head")
                    db.session.commit()
                    LOG.debug('Done.')
                    return True

                return True

        except sqlalchemy.exc.SQLAlchemyError as alch_err:
            LOG.error(str(alch_err))
            import traceback
            traceback.print_exc()
            return False

        except Exception as ex:
            LOG.error("Failed to create initial database schema")
            LOG.error(ex)
            import traceback
            traceback.print_exc()
            return False

    def get_schema_version(self):
        """
        Return the schema version from the database or a
        database status error code.
        """
        try:
            with DBContext(self) as db:
                if db.error:
                    return db.error

                mcontext = migration.MigrationContext.configure(db.connection)
                database_schema_revision = mcontext.get_current_revision()
                LOG.debug("Schema revision in the database: %s",
                          str(database_schema_revision))

                if database_schema_revision is None:
                    LOG.debug("Database schema should have been created!")
                    return DBStatus.SCHEMA_MISSING

                return database_schema_revision

        except sqlalchemy.exc.SQLAlchemyError as alch_err:
            LOG.debug(str(alch_err))
            return DBStatus.FAILED_TO_CONNECT

    def check_schema(self):
        """
        Checks the database schema for schema mismatch.
        The database server has to be started before this method is called.
        """

        try:
            with DBContext(self) as db:
                if db.error:
                    return db.error

                cfg = config.Config()
                cfg.set_main_option("script_location", self.migration_root)
                cfg.attributes["connection"] = db.connection

                scriptt = script.ScriptDirectory.from_config(cfg)
                mcontext = migration.MigrationContext.configure(
                    db.connection)
                database_schema_revision = mcontext.get_current_revision()
                LOG.debug("Schema revision in the database: %s",
                          str(database_schema_revision))

                if database_schema_revision is None:
                    LOG.debug("Database schema should have been created!")
                    return DBStatus.SCHEMA_MISSING

                LOG.debug("Checking schema versions in the package.")
                schema_config_head = scriptt.get_current_head()

                if database_schema_revision != schema_config_head:
                    LOG.debug("Database schema mismatch detected "
                              "between the package and the database")
                    LOG.debug("Checking if automatic upgrade is possible.")
                    all_revs = [rev.revision for rev in
                                scriptt.walk_revisions()]

                    if database_schema_revision not in all_revs:
                        LOG.debug("Automatic schema upgrade is not possible!")
                        LOG.debug("Please re-check your database and"
                                  "CodeChecker versions!")
                        return DBStatus.SCHEMA_MISMATCH_NO

                    # There is a schema mismatch.
                    return DBStatus.SCHEMA_MISMATCH_OK
                else:
                    LOG.debug("Schema in the package and in the database "
                              "is the same.")
                    LOG.debug("No schema modification is needed.")
                    return DBStatus.OK

        except sqlalchemy.exc.SQLAlchemyError as alch_err:
            LOG.error(str(alch_err))
            import traceback
            traceback.print_exc()
            return DBStatus.FAILED_TO_CONNECT
        except CommandError as cerr:
            LOG.error("Database schema and CodeChecker is incompatible. "
                      "Please update CodeChecker.")
            LOG.error(str(cerr))
            import traceback
            traceback.print_exc()
            return DBStatus.SCHEMA_MISMATCH_NO

    def upgrade(self):
        """
        Upgrade database schema.

        Checks the database schema for schema mismatch.
        The database server has to be started before this method is called.

        This method runs an Alembic upgrade to HEAD.
        """

        # Another safety check before we initialize or upgrade the schema.
        ret = self.check_schema()
        if ret not in [DBStatus.SCHEMA_MISMATCH_OK, DBStatus.SCHEMA_MISSING]:
            # Schema migration is not possible.
            return ret

        try:
            with DBContext(self) as db:
                if db.error:
                    return db.error

                LOG.debug("Create/update database schema for %s",
                          self.__model_meta['identifier'])

                cfg = config.Config()
                cfg.set_main_option("script_location", self.migration_root)
                cfg.attributes["connection"] = db.connection
                cfg.attributes["database_name"] = self.name_in_log

                command.upgrade(cfg, "head")
                db.session.commit()

                LOG.debug("Upgrading database schema: Done.")
                return DBStatus.OK

        except sqlalchemy.exc.SQLAlchemyError as alch_err:
            import traceback
            traceback.print_exc()
            LOG.error(str(alch_err))
            return DBStatus.SCHEMA_UPGRADE_FAILED

        except CommandError as cerr:
            import traceback
            traceback.print_exc()
            LOG.error(str(cerr))
            return DBStatus.SCHEMA_UPGRADE_FAILED

    @abstractmethod
    def connect(self, init=False):
        """
        Starts the database server and initializes the database server.

        On init == True, this it also initializes the database data and schema
        if needed.

        """
        pass

    @abstractmethod
    def get_connection_string(self) -> str:
        """
        Returns the connection string for SQLAlchemy.

        DO NOT LOG THE CONNECTION STRING BECAUSE IT MAY CONTAIN THE PASSWORD
        FOR THE DATABASE!
        """
        raise NotImplementedError()

    @abstractmethod
    def get_db_location(self):
        """
        Returns the database location.

        DATABASE USERNAME AND PASSWORD SHOULD NOT BE RETURNED HERE!
        """
        pass

    def get_model_identifier(self):
        return self.__model_meta['identifier']

    def _register_engine_hooks(self, engine):
        """
        This method registers hooks, if needed, related to the engine created
        by create_engine.
        """
        pass

    def create_engine(self):
        """
        Creates a new SQLAlchemy engine.
        """

        if make_url(self.get_connection_string()).drivername == \
                'sqlite+pysqlite':
            # FIXME: workaround for locking errors
            # FIXME: why is the connection used by multiple threads
            # is that a problem ??? do we need some extra locking???
            engine = sqlalchemy.create_engine(self.get_connection_string(),
                                              encoding='utf8',
                                              connect_args={'timeout': 600,
                                              'check_same_thread': False},
                                              poolclass=NullPool)
        else:
            engine = sqlalchemy.create_engine(self.get_connection_string(),
                                              encoding='utf8',
                                              poolclass=NullPool)

        self._register_engine_hooks(engine)
        return engine

    @staticmethod
    def connection_string_to_args(connection_string):
        """
        Convert the given connection string to an argument dict.

        !CAREFUL! This dict MIGHT contain a database password which SHOULD NOT
        be given to users!
        """

        url = make_url(connection_string)

        # Create an args for from_cmdline_args.
        args = {}
        if 'postgresql' in url.drivername:
            args['postgresql'] = True
            args['dbaddress'] = url.host
            args['dbport'] = url.port
            args['dbusername'] = url.username
            args['dbpassword'] = url.password
            args['dbname'] = url.database
        elif 'sqlite' in url.drivername:
            args['postgresql'] = False
            args['sqlite'] = url.database

        return args

    @staticmethod
    def from_connection_string(connection_string: str,
                               name_in_log: str,
                               model_meta,
                               migration_root,
                               interactive=False,
                               env=None):
        """
        Normally only this method is called form outside of this module in
        order to instance the proper server implementation.

        Parameters:
            connection_string: A fully formatted database URN identifying the
                connection parameters.
            name_in_log: A user-facing name of the current database to appear
                in log output.
            model_meta: the meta identifier of the database model to use
            migration_root: path to the database migration scripts
            env: a run environment dictionary.
        """

        args = SQLServer.connection_string_to_args(connection_string)
        return SQLServer.from_cmdline_args(args,
                                           name_in_log,
                                           model_meta,
                                           migration_root,
                                           interactive,
                                           env)

    @staticmethod
    def from_cmdline_args(args, name_in_log: str, model_meta, migration_root,
                          interactive=False, env=None):
        """
        Normally only this method is called form outside of this module in
        order to instance the proper server implementation.

        Parameters:
            args: the command line arguments from CodeChecker.py, but as a
              dictionary (if argparse.Namespace, use vars(args)).
            name_in_log: A custom user-friendly identifier for the DB in
              logs.
            model_meta: the meta identifier of the database model to use
            migration_root: path to the database migration scripts
            interactive: whether or not the database connection can be
              interactive on the server's shell.
            env: a run environment dictionary.
        """

        if not host_check.check_sql_driver(args['postgresql']):
            LOG.error("The selected SQL driver is not available!")
            raise IOError("The SQL driver to be used is not available!")

        if args['postgresql']:
            LOG.debug("Using PostgreSQL:")
            return PostgreSQLServer(name_in_log,
                                    model_meta,
                                    migration_root,
                                    args['dbaddress'],
                                    args['dbport'],
                                    args['dbusername'],
                                    args['dbname'],
                                    password=args['dbpassword']
                                    if 'dbpassword' in args else None,
                                    interactive=interactive,
                                    run_env=env)
        else:
            LOG.debug("Using SQLite:")
            data_file = os.path.abspath(args['sqlite'])
            LOG.debug("Database at %s", data_file)
            return SQLiteDatabase(name_in_log,
                                  data_file,
                                  model_meta,
                                  migration_root,
                                  run_env=env)


class PostgreSQLServer(SQLServer):
    """
    Handler for PostgreSQL.
    """

    def __init__(self, name_in_log, model_meta, migration_root,
                 host, port, user, database, password=None,
                 interactive=False, run_env=None):
        super().__init__(name_in_log, model_meta, migration_root)

        self.host = host
        self.port = port
        self.user = user
        self.database = database
        self.password = password
        self.interactive = interactive
        self.run_env = run_env

    def _get_connection_string(self, database):
        """
        Helper method for getting the connection string for the given database.

        database -- The user can force the database name in the returning
        connection string. However the password, if any, provided e.g. in a
        .pgpass file will be queried based on the database name which is given
        as a command line argument, even if it has a default value. The reason
        is that sometimes a connection with a common database name is needed,
        (e.g. 'postgres'), which requires less user permission.
        """

        driver = host_check.get_postgresql_driver_name()
        password = self.password
        if driver == 'pg8000' and not password:
            pfilepath = os.environ.get('PGPASSFILE')
            if pfilepath:
                password = pgpass.get_password_from_file(pfilepath,
                                                         self.host,
                                                         str(self.port),
                                                         self.database,
                                                         self.user)

        extra_args = {}
        if driver == "psycopg2":
            extra_args = {'client_encoding': 'utf8'}
        return str(URL('postgresql+' + driver,
                       username=self.user,
                       password=password,
                       host=self.host,
                       port=str(self.port),
                       database=database,
                       query=extra_args))

    def connect(self, init=False):
        """
        Connect to a PostgreSQL instance with given path, host and port.
        """

        LOG.debug("Connecting to database...")

        LOG.debug("Checking if database is running at [%s:%s]",
                  self.host, str(self.port))

        if self.user:
            # Try to connect to a specific database
            # with a specific user.
            check_db = ['psql',
                        '-h', self.host,
                        '-p', str(self.port),
                        '-U', self.user,
                        '-d', self.database,
                        '-c', 'SELECT version();']
        else:
            # Try to connect with the default settings.
            check_db = ['psql',
                        '-h', self.host,
                        '-p', str(self.port),
                        '-c', 'SELECT version();']

        if not self.interactive:
            # Do not prompt for password in non-interactive mode.
            check_db.append('--no-password')

        # If the user has a password pre-specified, use that for the
        # 'psql' call!
        env = self.run_env if self.run_env else os.environ
        env = env.copy()
        if self.password:
            env['PGPASSWORD'] = self.password

        err, code = call_command(check_db, env)

        if code:
            LOG.debug(err)
            return DBStatus.FAILED_TO_CONNECT

        if init:
            if not self._create_schema():
                return DBStatus.SCHEMA_INIT_ERROR

        return self.check_schema()

    def get_connection_string(self) -> str:
        return self._get_connection_string(self.database)

    def get_db_location(self):
        return self.host + ":" + str(self.port) + "/" + self.database


class SQLiteDatabase(SQLServer):
    """
    Handler for SQLite.
    """

    def __init__(self, name_in_log, data_file, model_meta, migration_root,
                 run_env=None):
        super().__init__(name_in_log, model_meta, migration_root)

        self.dbpath = data_file
        self.run_env = run_env

    def _register_engine_hooks(self, engine):
        """
        SQLite databases need FOREIGN KEYs to be enabled, which is handled
        through this connection hook.
        """
        def _set_sqlite_pragma(dbapi_connection, connection_record):
            cursor = dbapi_connection.cursor()
            cursor.execute("PRAGMA foreign_keys=ON")
            cursor.close()

        event.listen(engine, 'connect', _set_sqlite_pragma)

    def connect(self, init=False):
        if init:
            if not self._create_schema():
                LOG.error("Failed to create schema")
                return DBStatus.SCHEMA_INIT_ERROR

        return self.check_schema()

    def get_connection_string(self) -> str:
        return str(URL('sqlite+pysqlite', None, None, None, None, self.dbpath))

    def get_db_location(self):
        return self.dbpath


def conv(filter_value):
    """
    Convert * to % got from clients for the database queries.
    """
    if filter_value is None:
        return '%'
    return filter_value.replace('*', '%')


def escape_like(string, escape_char='*'):
    """Escape the string parameter used in SQL LIKE expressions."""
    return string.replace(escape_char, escape_char * 2) \
                 .replace('%', escape_char + '%') \
                 .replace('_', escape_char + '_')



"""
Disable review_status change

Revision ID: 3335ff7593cc
Revises:     4964142b58d2
Create Date: 2018-11-29 14:16:58.170551
"""

from alembic import op
import sqlalchemy as sa

# Revision identifiers, used by Alembic.
revision = '3335ff7593cc'
down_revision = '4964142b58d2'
branch_labels = None
depends_on = None


def upgrade():
    ctx = op.get_context()
    dialect = ctx.dialect.name

    new_col = sa.Column('is_review_status_change_disabled',
                        sa.Boolean(),
                        server_default=sa.sql.false())

    # To eliminate 'Skipping unsupported ALTER for creation of implicit
    # constraint' warning we use batch operation to create the new column in
    # case of SQLite.
    if dialect == 'sqlite':
        with op.batch_alter_table('products') as batch_op:
            batch_op.add_column(new_col)
    else:
        op.add_column('products', new_col)


def downgrade():
    op.drop_column('products', 'is_review_status_change_disabled')



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
"""
Result handlers to manage the output of the static analyzers.
"""

import hashlib
import os
import shlex

from abc import ABCMeta
from typing import Optional

from codechecker_analyzer import analyzer_context
from codechecker_common.logger import get_logger
from codechecker_common.skiplist_handler import SkipListHandlers
from codechecker_common.review_status_handler import ReviewStatusHandler


LOG = get_logger('analyzer')


class ResultHandler(metaclass=ABCMeta):
    """
    Handle and store the results at runtime for the analyzer:
    stdout, stderr, temporarily generated files.
    Do result postprocessing if required.

    For each buildaction there is one result handler.
    The result handler can handle multiple results if there
    are more than one analyzed source file in one buildaction.

    handle_results() handles the results and returns report statistics.

    For each build action
    - postprocess_result and handle_results can be called multiple times
      for the source files in one buildaction they will be analyzed separately
      with the same build options.

    Method call order should be:
    postprocess_result()
    handle_results()
    """
    # Handle the output stdout, or plist or both for an analyzer.

    def __init__(self, action, workspace, report_hash_type=None):
        """
        Put the temporary files for the workspace.
        """
        self.__workspace = workspace

        self.analyzer_cmd = []
        self.analyzer_stdout = ''
        self.analyzer_stderr = ''
        self.checker_labels = analyzer_context.get_context().checker_labels
        self.skiplist_handler = None
        self.analyzed_source_file = None
        self.analyzer_returncode = 1
        self.buildaction_hash = ''
        self.__buildaction = action

        self.__result_file = None
        self.__fixit_file = None

        # Report hash type can influence the post processing
        # of the results by rewriting the generated
        # report id (hash) values.
        self.report_hash_type = report_hash_type

    @property
    def buildaction(self):
        """
        """
        return self.__buildaction

    @property
    def source_dir_path(self):
        """ Get directory path of the compiled source file. """
        return os.path.normpath(os.path.join(
            os.getcwd(), self.__buildaction.directory))

    @property
    def workspace(self):
        """
        Workspace where the analysis results and temporarily generated files
        should go.
        """
        return self.__workspace

    @property
    def analyzer_action_str(self):
        """
        Generate a string which is unique to the analyzed build action. The
        returned string contains information of the analyzed file, the analyzer
        and the build command.
        """
        analyzed_file_name = os.path.basename(self.analyzed_source_file)

        source_file = os.path.normpath(
            os.path.join(self.buildaction.directory,
                         self.analyzed_source_file))

        # In case of "make 4.3" depending on compile-time options "make" tool
        # can be built so a subprocess named cc1build will be logged by
        # "CodeChecker log".
        # See posix_spawn() option:
        # https://lists.gnu.org/archive/html/info-gnu/2020-01/msg00004.html
        # In this case the -o output argument of cc1build command is a randomly
        # named temporary file. We can't afford dynamic parts in the original
        # build command, because its hash is used for identification in the
        # plist file name.
        #
        # The proper logging of this "make 4.3" version has been done in
        # bf140d6, so it is unlikely happen that two build actions differ only
        # in their "-o" flags. This workaround is still kept for any case.
        #
        # Note that some information loss occurs during the following algorithm
        # because ' '.join(shlex.split(cmd)) is not necessarily equal to cmd:
        # g++ -DVAR="hello world" main.cpp

        args = shlex.split(self.buildaction.original_command)
        indices = [idx for idx, v in enumerate(args) if v.startswith('-o')]

        for idx in reversed(indices):
            # Output can be given separate or joint:
            # -o a.out vs. -oa.out
            # In the first case we delete its argument too.
            if args[idx] == '-o':
                del args[idx]
            del args[idx]

        build_info = source_file + '_' + ' '.join(args)

        self.buildaction_hash = \
            hashlib.md5(build_info.encode(errors='ignore')).hexdigest()

        return analyzed_file_name + '_' + \
            str(self.buildaction.analyzer_type) + '_' + \
            self.buildaction_hash

    @property
    def analyzer_result_file(self):
        """
        Generate a result filename where the analyzer should put the results.
        Result file should be removed by the result handler eventually.
        """
        if not self.__result_file:
            self.__result_file = os.path.join(
                self.__workspace,
                self.analyzer_action_str + '.plist')

        return self.__result_file

    @property
    def fixit_file(self):
        """
        Generate a filename where the analyzer should put the fixit results.
        This is a .yaml file which contains the replacements which can be
        applied by clang-apply-replacements tool.
        """
        if not self.__fixit_file:
            self.__fixit_file = os.path.join(
                self.__workspace,
                'fixit',
                self.analyzer_action_str + '.yaml')

        return self.__fixit_file

    def clean_results(self):
        """
        Should be called after the postprocessing and result handling is done.
        """
        if self.__result_file:
            try:
                os.remove(self.__result_file)
            except OSError as oserr:
                # There might be no result file if analysis failed.
                LOG.debug(oserr)

    # TODO: If the parameters are not optional then we can get rid of some
    #   extra checks.
    def postprocess_result(
        self,
        skip_handlers: Optional[SkipListHandlers],
        rs_handler: Optional[ReviewStatusHandler]
    ):
        """
        Postprocess result if needed.
        Should be called after the analyses finished.
        """
        pass

    def handle_results(self, client):
        """
        Handle the results and return report statistics.
        """
        pass



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
"""
Prepare and start different analysis types
"""


from collections import defaultdict
import os
import shutil
import signal
import sys
import time

from multiprocess.managers import SyncManager

from codechecker_common.logger import get_logger
from codechecker_common.review_status_handler import ReviewStatusHandler

from . import analyzer_context, analysis_manager, pre_analysis_manager, \
    checkers
from .analyzers import analyzer_types
from .analyzers.config_handler import AnalyzerConfigHandler, CheckerState
from .analyzers.clangsa.analyzer import ClangSA

from .makefile import MakeFileCreator

LOG = get_logger('analyzer')


def prepare_actions(actions, enabled_analyzers):
    """
    Set the analyzer type for each buildaction.
    Multiple actions if multiple source analyzers are set.
    """
    res = []

    for ea in enabled_analyzers:
        for action in actions:
            res.append(action.with_attr('analyzer_type', ea))
    return res


def create_actions_map(actions, manager):
    """
    Create a dict for the build actions which is shareable
    safely between processes.
    Key: (source_file, target)
    Value: BuildAction
    """

    result = manager.dict()

    for act in actions:
        key = act.source, act.target
        if key in result:
            LOG.debug("Multiple entires in compile database "
                      "with the same (source, target) pair: (%s, %s)",
                      act.source, act.target)
        result[key] = act
    return result


def __mgr_init():
    """
    This function is set for the SyncManager object which handles shared data
    structures among the processes of the pool. Ignoring the SIGINT signal is
    necessary in the manager object so it doesn't terminate before the
    termination of the process pool.
    """
    signal.signal(signal.SIGINT, signal.SIG_IGN)


def __get_statistics_data(args):
    """ Get statistics data. """
    statistics_data = None

    if 'stats_enabled' in args and args.stats_enabled:
        statistics_data = {
            'stats_out_dir': os.path.join(args.output_path, "stats")}

    if 'stats_output' in args and args.stats_output:
        statistics_data = {'stats_out_dir': args.stats_output}

    if statistics_data:
        statistics_data['stat_tmp_dir'] = \
            os.path.join(statistics_data.get('stats_out_dir'), 'tmp')

    if 'stats_min_sample_count' in args and statistics_data:
        if args.stats_min_sample_count > 1:
            statistics_data['stats_min_sample_count'] =\
                args.stats_min_sample_count
        else:
            LOG.error("stats_min_sample_count"
                      "must be greater than 1.")
            return None

    if 'stats_relevance_threshold' in args and statistics_data:
        if 1 > args.stats_relevance_threshold > 0:
            statistics_data['stats_relevance_threshold'] =\
                args.stats_relevance_threshold
        else:
            LOG.error("stats-relevance-threshold must be"
                      " greater than 0 and smaller than 1.")
            return None

    return statistics_data


def __get_ctu_data(ctu_dir):
    """ Get CTU data. """
    ctu_capability = ClangSA.ctu_capability()
    return {'ctu_dir': ctu_dir,
            'ctu_func_map_cmd': ctu_capability.mapping_tool_path,
            'ctu_func_map_file': ctu_capability.mapping_file_name,
            'ctu_temp_fnmap_folder': 'tmpExternalFnMaps'}


def __has_enabled_checker(ch: AnalyzerConfigHandler):
    """
    Returns True if at least one checker is enabled in the given config
    handler.
    """
    return any(state == CheckerState.enabled
               for _, (state, _) in ch.checks().items())


def perform_analysis(args, skip_handlers, rs_handler: ReviewStatusHandler,
                     actions, metadata_tool, compile_cmd_count):
    """
    Perform static analysis via the given (or if not, all) analyzers,
    in the given analysis context for the supplied build actions.
    Additionally, insert statistical information into the metadata dict.
    """
    # pylint: disable=no-member multiprocess module members.

    context = analyzer_context.get_context()

    ctu_reanalyze_on_failure = 'ctu_reanalyze_on_failure' in args and \
        args.ctu_reanalyze_on_failure

    analyzers = args.analyzers if 'analyzers' in args \
        else analyzer_types.supported_analyzers
    analyzers, errored = analyzer_types.check_supported_analyzers(analyzers)
    analyzer_types.check_available_analyzers(analyzers, errored)

    ctu_collect = False
    ctu_analyze = False
    ctu_dir = ''
    if 'ctu_phases' in args:
        ctu_dir = os.path.join(args.output_path, 'ctu-dir')
        args.ctu_dir = ctu_dir
        if ClangSA.ANALYZER_NAME not in analyzers:
            LOG.error("CTU can only be used with the clang static analyzer.")
            return
        ctu_collect = args.ctu_phases[0]
        ctu_analyze = args.ctu_phases[1]

    if 'stats_enabled' in args and args.stats_enabled:
        if ClangSA.ANALYZER_NAME not in analyzers:
            LOG.debug("Statistics can only be used with "
                      "the Clang Static Analyzer.")
            return

    config_map = analyzer_types.build_config_handlers(args, analyzers)

    # Don't enable analyzers that have no checkers enabled. Some analyzers,
    # like clang-tidy error out if none of theirs are enabled, which makes
    # sense if you use them as a standalone tool, but not if you run a bunch
    # of analyzers at once.
    no_checker_analyzers = \
        [a for a in analyzers if not __has_enabled_checker(config_map[a])]
    for analyzer in no_checker_analyzers:
        LOG.info("No checkers enabled for %s", analyzer)
        analyzers.remove(analyzer)

    # Don't emit warnings for analyzers we failed to enable if none of their
    # checkers were enabled anyways.
    # TODO: Its perfectly reasonable for an analyzer plugin to not be able to
    # build their config handler if the analyzer isn't supported in the first
    # place. For now, this seems to be okay, but may not be later.
    # Even then, if we couldn't get hold of the analyzer binary, we can't do
    # anything.
    errored_config_map = analyzer_types.build_config_handlers(
        args,
        [x for (x, _) in errored
         if analyzer_types.supported_analyzers[x].analyzer_binary()])

    no_checker_err_analyzers = \
        [analyzer for analyzer, config_h in errored_config_map.items()
         if not __has_enabled_checker(config_h)]

    errored = [(an, msg) for an, msg in errored
               if an not in no_checker_err_analyzers]

    actions = prepare_actions(actions, analyzers)

    available_checkers = set()
    # Add profile names to the checkers list so we will not warn
    # if a profile is enabled but there is no checker with that name.

    available_checkers.update(
        context.checker_labels.get_description('profile'))

    # Collect all the available checkers from the enabled analyzers.
    for analyzer in config_map.items():
        _, analyzer_cfg = analyzer
        for analyzer_checker in analyzer_cfg.checks().items():
            checker_name, _ = analyzer_checker
            available_checkers.add(checker_name)

    if 'ordered_checkers' in args:
        missing_checkers = checkers.available(args.ordered_checkers,
                                              available_checkers)
        if missing_checkers:
            diag_msg = "No checker(s) with these names was found:\n{}".format(
                '\n'.join(missing_checkers))
            if 'no_missing_checker_error' in args:
                LOG.warning(diag_msg)
            else:
                LOG.error(diag_msg)
                LOG.info("Although it is not reccomended, if you want to "
                         "suppress errors relating to unknown "
                         "checker names, consider using the option "
                         "'--no-missing-checker-error'")
                sys.exit(1)

    enabled_checkers = defaultdict(list)

    # Save some metadata information.
    for analyzer in analyzers:
        metadata_info = {
            'checkers': {},
            'analyzer_statistics': {
                "failed": 0,
                "failed_sources": [],
                "successful": 0,
                "successful_sources": [],
                "version": None}}

        for check, data in config_map[analyzer].checks().items():
            state, _ = data
            metadata_info['checkers'].update({
                check: state == CheckerState.enabled})
            if state == CheckerState.enabled:
                enabled_checkers[analyzer].append(check)

        # TODO: cppcheck may require a different environment than clang.
        version = analyzer_types.supported_analyzers[analyzer] \
            .get_binary_version(context.analyzer_env)
        metadata_info['analyzer_statistics']['version'] = version

        metadata_tool['analyzers'][analyzer] = metadata_info
    LOG.info("Enabled checker list can be found in %s",
             os.path.join(args.output_path, "metadata.json"))
    LOG.debug("Enabled checkers:\n%s", '\n'.join(
              k + ': ' + ',\n '.join(v) for k, v in enabled_checkers.items()))

    if 'makefile' in args and args.makefile:
        statistics_data = __get_statistics_data(args)

        ctu_data = None
        if ctu_collect or statistics_data:
            ctu_data = __get_ctu_data(ctu_dir)

        makefile_creator = MakeFileCreator(analyzers, args.output_path,
                                           config_map, skip_handlers,
                                           ctu_collect, statistics_data,
                                           ctu_data)
        makefile_creator.create(actions)
        return

    if ctu_collect:
        shutil.rmtree(ctu_dir, ignore_errors=True)
    elif ctu_analyze and not os.path.exists(ctu_dir):
        LOG.error("CTU directory: '%s' does not exist.", ctu_dir)
        return

    start_time = time.time()

    # Use Manager to create data objects which can be
    # safely shared between processes.
    manager = SyncManager()
    manager.start(__mgr_init)

    config_map = manager.dict(config_map)
    actions_map = create_actions_map(actions, manager)

    # Setting to not None value will enable statistical analysis features.
    statistics_data = __get_statistics_data(args)
    if statistics_data:
        statistics_data = manager.dict(statistics_data)

    if ctu_collect or statistics_data:
        ctu_data = None
        if ctu_collect or ctu_analyze:
            ctu_data = manager.dict(__get_ctu_data(ctu_dir))

        pre_analyze = [a for a in actions
                       if a.analyzer_type == ClangSA.ANALYZER_NAME]
        pre_anal_skip_handlers = None

        # Skip list is applied only in pre-analysis
        # if --ctu-collect or --stats-collect  was called explicitly
        if ((ctu_collect and not ctu_analyze)
                or ("stats_output" in args and args.stats_output)):
            pre_anal_skip_handlers = skip_handlers

        clangsa_config = config_map.get(ClangSA.ANALYZER_NAME)

        if clangsa_config is not None:
            pre_analysis_manager.run_pre_analysis(pre_analyze,
                                                  clangsa_config,
                                                  args.jobs,
                                                  pre_anal_skip_handlers,
                                                  ctu_data,
                                                  statistics_data,
                                                  manager)
        else:
            LOG.error("Can not run pre analysis without clang "
                      "static analyzer configuration.")

    if 'stats_output' in args and args.stats_output:
        return

    if 'stats_dir' in args and args.stats_dir:
        statistics_data = manager.dict({'stats_out_dir': args.stats_dir})

    if ctu_analyze or statistics_data or (not ctu_analyze and not ctu_collect):

        LOG.info("Starting static analysis ...")
        analysis_manager.start_workers(actions_map, actions,
                                       config_map, args.jobs,
                                       args.output_path,
                                       skip_handlers,
                                       rs_handler,
                                       metadata_tool,
                                       'quiet' in args,
                                       'capture_analysis_output' in args,
                                       'generate_reproducer' in args,
                                       args.timeout if 'timeout' in args
                                       else None,
                                       ctu_reanalyze_on_failure,
                                       statistics_data,
                                       manager,
                                       compile_cmd_count)
        LOG.info("Analysis finished.")
        LOG.info("To view results in the terminal use the "
                 "\"CodeChecker parse\" command.")
        LOG.info("To store results use the \"CodeChecker store\" command.")
        LOG.info("See --help and the user guide for further options about"
                 " parsing and storing the reports.")
        LOG.info("----=================----")

    end_time = time.time()
    LOG.info("Analysis length: %s sec.", end_time - start_time)

    analyzer_types.print_unsupported_analyzers(errored)

    metadata_tool['timestamps'] = {'begin': start_time,
                                   'end': end_time}

    if ctu_collect and ctu_analyze:
        shutil.rmtree(ctu_dir, ignore_errors=True)

    manager.shutdown()



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
"""
Defines the CodeChecker action for applying fixits.
"""

import argparse
import json
import os
import re
import subprocess
import sys
import tempfile
import yaml

from codechecker_report_converter.util import get_last_mod_time

from codechecker_analyzer import analyzer_context
from codechecker_analyzer.analyzers import analyzer_types
from codechecker_common import arg, logger

LOG = logger.get_logger('system')


def get_argparser_ctor_args():
    """
    This method returns a dict containing the kwargs for constructing an
    argparse.ArgumentParser (either directly or as a subparser).
    """

    return {
        'prog': 'CodeChecker fixit',
        'formatter_class': arg.RawDescriptionDefaultHelpFormatter,
        'description': """
Some analyzers may suggest some automatic bugfixes. Most of the times these are
style issues which can be fixed easily. This command handles the listing and
application of these automatic fixes.

Besides the provided filter options you can pipe the JSON format output of
"CodeChecker cmd diff" command to filter automatic fixes only on new reports:
CodeChecker cmd diff -b dir1 -n dir2 -o json --new | CodeChecker fixit dir2""",
        'help': "Apply automatic fixes based on the suggestions of the "
                "analyzers"
    }


def add_arguments_to_parser(parser):
    """
    Add the subcommand's arguments to the given argparse.ArgumentParser.
    """

    parser.add_argument('input',
                        type=str,
                        nargs='+',
                        metavar='folder',
                        help="The analysis result folder(s) containing "
                             "analysis results and fixits which should be "
                             "applied.")
    parser.add_argument('-a', '--apply',
                        action='store_true',
                        default=argparse.SUPPRESS,
                        help="Apply the available automatic fixes. This "
                             "causes the modification of the source code.")
    parser.add_argument('-i', '--interactive',
                        action='store_true',
                        default=False,
                        help="Interactive selection of fixits to apply. Fixit "
                             "items are enumerated one by one and you may "
                             "choose which ones are to be applied. This "
                             "causes the modification of the source code.")
    parser.add_argument('--checker-name',
                        nargs='*',
                        default=[],
                        help='Filter fixits by checker names. The checker '
                             'name can contain multiple * quantifiers which '
                             'matches any number of characters (zero or '
                             'more). So for example "*DeadStores" will '
                             'match "deadcode.DeadStores".')
    parser.add_argument('--file',
                        metavar='FILE_PATH',
                        nargs='*',
                        default=[],
                        help='Filter fixits by file path. The file path can '
                             'contain multiple * quantifiers which matches '
                             'any number of characters (zero or more). So if '
                             'you have /a/x.cpp and /a/y.cpp then "/a/*.cpp" '
                             'selects both.')

    logger.add_verbose_arguments(parser)
    parser.set_defaults(func=main)


def get_location_by_offset(filename, offset):
    """
    This function returns the line and column number in the given file which
    is located at the given offset (i.e. number of characters including new
    line characters).
    """
    with open(filename, encoding='utf-8', errors='ignore') as f:
        for row, line in enumerate(f, 1):
            length = len(line)
            if length < offset:
                offset -= length
            else:
                return row, offset + 1


def clang_tidy_fixit_filter(content, checker_names, file_paths, reports,
                            modification_time, interactive):
    """
    This function filters the content of a replacement .yaml file.
    The function returns the set of not-existing files which to which the
    fixits couldn't apply, and the set of changed files and the set of files
    modified after the given timestamp.
    content -- The content of a replacement .yaml file parsed to an object by
               yaml module.
    checker_names -- A list of checker names possibly containing * joker
                     characters. The full checker name must match in order to
                     apply it.
    file_paths -- A list of file paths to which the fixits will be applied.
                  A file path may possibly contain * joker characters. The
                  full path must match in order to apply it.
    reports -- A list of CodeChecker reports. This can come from
               "CodeChecker cmd [diff|results] ..." command in JSON format.
    modification_time -- Timestamp for the generation of fixit file. If a
                         source file has been modified after then that will be
                         filtered out.
    interactive -- Interactive filtering. If True then user will be asked about
                   each fixit one by one.
    """
    def make_regex(parts):
        """
        Transform a list of strings to a regex which matches any of the
        complete strings possibly containing * joker character.
        """
        if not parts:
            return re.compile('.*')
        parts = map(lambda part: re.escape(part).replace(r'\*', '.*'), parts)
        return re.compile('|'.join(parts) + '$')

    def ask_user(item):
        """
        Filter function for interactive filtering based on user input.
        """
        print(yaml.dump(item))
        prompt = "y/<Enter> - Yes\nOther     - No\nCtrl-C    - Cancel\nApply: "
        return input(prompt) in ('Y', 'y', '')

    def check_existence_and_collect(file_path):
        """
        Return True if the file exists otherwise collect it to a set named
        non_existing_files.
        """
        if not os.path.isfile(file_path):
            not_existing_files.add(file_path)
            return False
        return True

    def check_modification_and_collect(file_path):
        """
        Return True if the file was not modified after the given timestamp
        otherwise collect it to a set named modified_files.
        """
        if get_last_mod_time(file_path) > modification_time:
            modified_files.add(file_path)
            return False
        return True

    def match_reports(diag_msg):
        """
        Check if the fixit belongs to a CodeChecker report from "reports"
        object. In case "reports" are not given, this function returns True.
        """
        if not reports:
            return True

        row, col = get_location_by_offset(diag_msg['FilePath'],
                                          int(diag_msg['FileOffset']))
        # TODO: There are different formats for "reports" object. These are
        # supposed to be marked with a version number in the JSON object. For
        # example in some "CodeChecker cmd diff" command output there is no
        # "checkedFile" attribute for a report, but the file path is stored in
        # another structure.
        return any(row == report['line'] and col == report['column'] and
                   diag_msg['FilePath'] == report['checkedFile']
                   for report in reports)

    not_existing_files = set()
    existing_files = set()
    modified_files = set()

    checker_names = make_regex(checker_names)
    file_paths = make_regex(file_paths)

    items = filter(
        lambda diag: checker_names.match(diag['DiagnosticName']) and
        len(diag['DiagnosticMessage']['Replacements']) != 0 and
        file_paths.match(diag['DiagnosticMessage']['FilePath']) and
        match_reports(diag['DiagnosticMessage']) and
        check_existence_and_collect(diag['DiagnosticMessage']['FilePath']) and
        check_modification_and_collect(diag['DiagnosticMessage']['FilePath']),
        content['Diagnostics'])

    if interactive:
        items = filter(ask_user, items)

    content['Diagnostics'] = list(items)

    for diag in content['Diagnostics']:
        existing_files.add(diag['DiagnosticMessage']['FilePath'])

    return existing_files, not_existing_files, modified_files


def list_fixits(inputs, checker_names, file_paths, interactive, reports):
    """
    This function dumps the .yaml files to the standard output like a "dry run"
    with the replacements. See clang_tidy_fixit_filter() for the documentation
    of the filter options.
    inputs -- A list of report directories which contains the fixit dumps in a
              subdirectory named "fixit".
    """
    not_existing_files = set()
    existing_files = set()
    modified_files = set()

    for i in inputs:
        fixit_dir = os.path.join(i, 'fixit')

        if not os.path.isdir(fixit_dir):
            LOG.info('No fixits in %s', i)
            continue

        for fixit_file in os.listdir(fixit_dir):
            fixit_file = os.path.join(fixit_dir, fixit_file)
            with open(fixit_file, encoding='utf-8', errors='ignore') as f:
                content = yaml.load(f, Loader=yaml.BaseLoader)
                fixit_mtime = get_last_mod_time(fixit_file)

                existing, not_existing, modified = clang_tidy_fixit_filter(
                    content, checker_names, file_paths, reports, fixit_mtime,
                    interactive)

                existing_files.update(existing)
                not_existing_files.update(not_existing)
                modified_files.update(modified)

            if content['Diagnostics']:
                print(yaml.dump(content))

    if existing_files:
        print("Updated files:\n{}".format(
            '\n'.join(sorted(existing_files))),
            file=sys.stderr)
    if not_existing_files:
        print("Not existing files:\n{}".format(
            '\n'.join(sorted(not_existing_files))),
            file=sys.stderr)
    if modified_files:
        print("Skipped files due to modification since last analysis:\n{}"
              .format('\n'.join(sorted(modified_files))),
              file=sys.stderr)


def apply_fixits(inputs, checker_names, file_paths, interactive, reports):
    """
    This function applies the replacements from the .yaml files. A Clang tool
    clang-apply-replacements is used which may or may not provide
    --ignore-insert-conflict flag depending on the Clang version. If this flag
    doesn't exist then the replacement process stops when a .yaml file doesn't
    apply for some reason. In this case each .yaml file needs to be applied one
    by one.
    inputs -- A list of report directories which contains the fixit dumps in a
              subdirectory named "fixit".
    """
    def apply_process(out_dir):
        """
        Execute clang-apply-replacements binary.
        """
        subprocess.Popen([
            analyzer_context.get_context().replacer_binary,
            *ignore_flag,
            out_dir]).communicate()

    not_existing_files = set()
    existing_files = set()
    modified_files = set()

    ignore_flag = ["--ignore-insert-conflict"] if \
        analyzer_types.is_ignore_conflict_supported() else []

    for i in inputs:
        fixit_dir = os.path.join(i, 'fixit')

        if not os.path.isdir(fixit_dir):
            LOG.info('No fixits in %s', i)
            continue

        with tempfile.TemporaryDirectory() as out_dir:
            for fixit_file in os.listdir(fixit_dir):
                with open(os.path.join(fixit_dir, fixit_file),
                          encoding='utf-8', errors='ignore') as f:
                    content = yaml.load(f, Loader=yaml.BaseLoader)
                    fixit_mtime = get_last_mod_time(
                        os.path.join(fixit_dir, fixit_file))

                    existing, not_existing, modified = clang_tidy_fixit_filter(
                        content, checker_names, file_paths, reports,
                        fixit_mtime, interactive)

                    existing_files.update(existing)
                    not_existing_files.update(not_existing)
                    modified_files.update(modified)

                fixit_file = os.path.join(out_dir, fixit_file)
                if len(content['Diagnostics']) != 0:
                    with open(fixit_file, 'w',
                              encoding='utf-8', errors='ignore') as out:
                        yaml.dump(content, out)

                if not ignore_flag:
                    apply_process(out_dir)
                    try:
                        os.remove(fixit_file)
                    except FileNotFoundError:
                        pass

            if ignore_flag:
                apply_process(out_dir)

    if existing_files:
        print("Updated files:\n{}".format(
            '\n'.join(sorted(existing_files))),
            file=sys.stderr)
    if not_existing_files:
        print("Not existing files:\n{}".format(
            '\n'.join(sorted(not_existing_files))),
            file=sys.stderr)
    if modified_files:
        print("Skipped files due to modification since last analysis:\n{}"
              .format('\n'.join(sorted(modified_files))),
              file=sys.stderr)


def main(args):
    """
    Entry point for the command handling automatic fixes.

    TODO: Currently clang-tidy is the only tool which supports the dumping of
    fixit replacements. In this script we assume that the replacement dump
    .yaml files are in the format so clang-apply-replacement Clang tool can
    consume them.
    """

    logger.setup_logger(args.verbose if 'verbose' in args else None)

    context = analyzer_context.get_context()

    if not context.replacer_binary:
        LOG.error("clang-apply-replacements tool is not found")
        return

    if not sys.stdin.isatty() and args.interactive:
        LOG.error("Interactive mode can't be used in case of JSON data "
                  "arriving from standard input.")
        sys.exit(1)

    try:
        reports = None if sys.stdin.isatty() else json.loads(sys.stdin.read())
    except json.decoder.JSONDecodeError as ex:
        LOG.error("JSON format error on standard input: %s", ex)
        sys.exit(1)

    # "CodeChecker fixit" can be applied on the output of
    # "CodeChecker cmd diff" command. Earlier this was a simple list of
    # reports, but now this is an object with the list of reports and a version
    # number indicating the version of the object's format.
    if reports is not None and 'version' in reports:
        reports = reports['reports']

    if 'apply' in args or args.interactive:
        apply_fixits(args.input, args.checker_name, args.file,
                     args.interactive, reports)
    else:
        list_fixits(args.input, args.checker_name, args.file,
                    args.interactive, reports)



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------

import logging
import re

from ..parser import SANParser

LOG = logging.getLogger('report-converter')


class Parser(SANParser):
    """ Parser for Clang LeakSanitizer console outputs. """

    checker_name = "LeakSanitizer"

    # Regex for parsing MemorySanitizer output message.
    line_re = re.compile(
        r'(?P<message>(Ind|D)irect leak of \d+ byte\(s\) '
        r'in \d+ object\(s\) allocated from:)')

    skip_line_after_match = False



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
"""Helper and converter functions for the gerrit review json format."""

import json
import logging
import os
import re

from typing import Dict, List, Union

from codechecker_report_converter.report import Report


LOG = logging.getLogger('report-converter')


def convert(reports: List[Report]) -> Dict:
    """Convert reports to gerrit review format.

    Process the required environment variables and convert the reports
    to the required gerrit json format.
    """
    repo_dir = os.environ.get('CC_REPO_DIR')
    report_url = os.environ.get('CC_REPORT_URL')
    changed_file_path = os.environ.get('CC_CHANGED_FILES')
    changed_files = __get_changed_files(changed_file_path)

    return __convert_reports(reports, repo_dir, report_url,
                             changed_files, changed_file_path)


def mandatory_env_var_is_set():
    """
    True if mandatory environment variables are set otherwise False and print
    error messages.
    """
    no_missing_env_var = True

    if os.environ.get('CC_REPO_DIR') is None:
        LOG.error("When using gerrit output the 'CC_REPO_DIR' environment "
                  "variable needs to be set to the root directory of the "
                  "sources, i.e. the directory where the repository was "
                  "cloned!")
        no_missing_env_var = False

    if os.environ.get('CC_CHANGED_FILES') is None:
        LOG.error("When using gerrit output the 'CC_CHANGED_FILES' "
                  "environment variable needs to be set to the path of "
                  "changed files json from Gerrit!")
        no_missing_env_var = False

    return no_missing_env_var


def __convert_reports(reports: List[Report],
                      repo_dir: Union[str, None],
                      report_url: Union[str, None],
                      changed_files: List[str],
                      changed_file_path: Union[str, None]) -> Dict:
    """Convert the given reports to gerrit json format.

    This function will convert the given report to Gerrit json format.
    reports - list of reports comming from a plist file or
              from the CodeChecker server (both types can be processed)
    repo_dir - Root directory of the sources, i.e. the directory where the
               repository was cloned.
    report_url - URL where the report can be found something like this:
      "http://jenkins_address/userContent/$JOB_NAME/$BUILD_NUM/index.html"
    changed_files - list of the changed files
    checker_labels
    """
    review_comments: Dict[str, List[Dict]] = {}

    report_count = 0
    report_messages_in_unchanged_files = []
    for report in reports:
        file_name = report.file.path

        report_count += 1

        # file_name can be without a path in the report.
        rel_file_path = os.path.relpath(file_name, repo_dir) \
            if repo_dir and os.path.dirname(file_name) != "" else file_name

        checked_file = rel_file_path \
            + ':' + str(report.line) + ":" + str(report.column)

        review_comment_msg = \
            f"[{report.severity}] {checked_file}: {report.message} " \
            f"[{report.checker_name}]\n{report.source_line}"

        # Skip the report if it is not in the changed files.
        if changed_file_path and not \
                any([file_name.endswith(c) for c in changed_files]):
            report_messages_in_unchanged_files.append(review_comment_msg)
            continue

        if rel_file_path not in review_comments:
            review_comments[rel_file_path] = []

        review_comments[rel_file_path].append({
            "range": {
                "start_line": report.line,
                "start_character": report.column,
                "end_line": report.line,
                "end_character": report.column},
            "message": review_comment_msg})

    message = f"CodeChecker found {report_count} issue(s) in the code."

    if report_messages_in_unchanged_files:
        message += ("\n\nThere following reports are introduced in files "
                    "which are not changed and can't be shown as individual "
                    "reports:\n{0}\n".format('\n'.join(
                        report_messages_in_unchanged_files)))

    if report_url:
        message += f" See: {report_url}"

    review = {"tag": "jenkins",
              "message": message,
              "labels": {
                  "Code-Review": -1 if report_count else 1,
                  "Verified": -1 if report_count else 1},
              "comments": review_comments}
    return review


def __get_changed_files(changed_file_path: Union[None, str]) -> List[str]:
    """Return a list of changed files.

    Process the given gerrit changed file object and return a list of
    file paths which changed.

    The file can contain some garbage values at start, so we use regex
    to find a json object.
    """
    changed_files: List[str] = []

    if not changed_file_path or not os.path.exists(changed_file_path):
        return changed_files

    with open(changed_file_path,
              encoding='utf-8',
              errors='ignore') as changed_file:
        content = changed_file.read()

        # The file can contain some garbage values at start, so we use
        # regex search to find a json object.
        match = re.search(r'\{[\s\S]*\}', content)
        if not match:
            return changed_files

        for filename in json.loads(match.group(0)):
            if "/COMMIT_MSG" in filename:
                continue

            changed_files.append(filename)

    return changed_files



"""
Add bug path length

Revision ID: e89887e7d3f0
Revises:     3793e361a752
Create Date: 2018-08-28 11:11:07.533906
"""

from alembic import op
import sqlalchemy as sa


# Revision identifiers, used by Alembic.
revision = 'e89887e7d3f0'
down_revision = '3793e361a752'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column('reports',
                  sa.Column('path_length', sa.Integer(),
                            nullable=True, default=0))

    conn = op.get_bind()

    conn.execute("""
        UPDATE reports
        SET path_length =
        (SELECT COUNT(bug_path_events.report_id)
            FROM bug_path_events
            WHERE bug_path_events.report_id = reports.id)
    """)


def downgrade():
    op.drop_column('reports', 'path_length')



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
"""
Static analyzer configuration handler.
"""


from abc import ABCMeta
from enum import Enum
import collections
import platform

from codechecker_analyzer import analyzer_context
from codechecker_common.logger import get_logger

LOG = get_logger('system')


# If the check appears in profiles, and such a profile is enabled explicitly on
# the command-line or implicitly as in case of the default profile, then they
# are considered to have a CheckerState of enabled. Likewise for individually
# enabled checks. If a check is however explicitly disabled on the
# command-line, or belongs to a profile explicitly disabled on the
# command-line, then it is considered to have a CheckerState of disabled.
class CheckerState(Enum):
    disabled = 1
    enabled = 2


class CheckerType(Enum):
    """
    Checker type.
    """
    analyzer = 0  # A checker which is not a compiler warning.
    compiler = 1  # A checker which specified as "-W<name>" or "-Wno-<name>".


def get_compiler_warning_name_and_type(checker_name):
    """
    Removes 'W' or 'Wno' from the compiler warning name, if this is a
    compiler warning and returns the name and CheckerType.compiler.
    If it is a clang-diagnostic-<name> warning then it returns the name
    and CheckerType.analyzer.
    Otherwise returns None and CheckerType.analyzer.
    """
    # Checker name is a compiler warning.
    if checker_name.startswith('W'):
        name = checker_name[4:] if \
            checker_name.startswith('Wno-') else checker_name[1:]
        return name, CheckerType.compiler
    elif checker_name.startswith('clang-diagnostic-'):
        return checker_name[17:], CheckerType.analyzer
    else:
        return None, CheckerType.analyzer


class AnalyzerConfigHandler(metaclass=ABCMeta):
    """
    Handle the checker configurations and enabled disabled checkers lists.
    """

    def __init__(self):

        self.analyzer_extra_arguments = []
        self.checker_config = ''
        self.analyzer_config = None
        self.report_hash = None
        self.enable_all = None

        # The key is the checker name, the value is a tuple of CheckerState and
        # checker description.
        self.__available_checkers = collections.OrderedDict()

    def add_checker(self, checker_name, description='',
                    state=CheckerState.disabled):
        """
        Add a checker to the available checkers' list.
        """
        self.__available_checkers[checker_name] = (state, description)

    def set_checker_enabled(self, checker_name, enabled=True):
        """
        Explicitly handle checker state, keep description if already set.
        """
        changed_states = []

        for ch_name, values in self.__available_checkers.items():
            if ch_name.startswith(checker_name) or \
               ch_name.endswith(checker_name):
                _, description = values
                state = CheckerState.enabled if enabled \
                    else CheckerState.disabled
                changed_states.append((ch_name, state, description))

        # Enabled/disable checkers are stored in an ordered dict. When a
        # checker group (e.g. clang-diagnostic) is disabled then all checkers
        # in this group are added to this list in order, including
        # clang-diagnostic itself at the end. Enabling a specific member of a
        # disabled group (e.g. clang-diagnostic-vla) resets the status at an
        # earlier position than the group name, so due to its order the group
        # disable will be stronger. The solution is removing the given checker
        # and inserting it back to the end.
        for ch_name, state, description in changed_states:
            del self.__available_checkers[ch_name]
            self.__available_checkers[ch_name] = (state, description)

    def checks(self):
        """
        Return the checkers.
        """
        return self.__available_checkers

    def __gen_name_variations(self):
        """
        Generate all applicable name variations from the given checker list.
        """
        checker_names = (name for name in self.__available_checkers)
        reserved_names = []

        for name in checker_names:
            delim = '.' if '.' in name else '-'
            parts = name.split(delim)
            # Creates a list of variations from a checker name, e.g.
            # ['security', 'security.insecureAPI', 'security.insecureAPI.gets']
            # from 'security.insecureAPI.gets' or
            # ['misc', 'misc-dangling', 'misc-dangling-handle']
            # from 'misc-dangling-handle'.
            v = [delim.join(parts[:(i + 1)]) for i in range(len(parts))]
            reserved_names += v

        return reserved_names

    def initialize_checkers(self,
                            checkers,
                            cmdline_enable=[],
                            enable_all=False):
        """
        Add checkers and set their "enabled/disabled" status. The following
        inputs are considered in this order:
        - First the default state is taken based on the analyzer tool.
        - Members of "default" profile are enabled.
        - In case of "--enable-all" every checker is enabled except for "alpha"
          and "debug" checker groups. "osx" checker group is also not included
          unless the target platform is Darwin.
        - Command line "--enable/--disable" flags.
          - Their arguments may start with "profile:" or "guideline:" prefix
            which makes the choice explicit.
          - Without prefix it means a profile name, a guideline name or a
            checker group/name in this priority order.

        checkers -- [(checker name, description), ...] Checkers to add with
                    their description.
        cmdline_enable -- [(argument, enabled), ...] Arguments of
                          "--enable/--disable" flags and a boolean value
                          whether it is after "--enable" or not.
        enable_all -- Boolean value whether "--enable-all" is given.
        """

        checker_labels = analyzer_context.get_context().checker_labels

        # Add all checkers marked as disabled.
        for checker_name, description in checkers:
            self.add_checker(checker_name, description)

        # Set default enabled or disabled checkers, based on the config file.
        default_profile_checkers = \
            checker_labels.checkers_by_labels(['profile:default'])
        if not default_profile_checkers:
            # Check whether a default profile exists.
            LOG.warning("No default profile found!")
        else:
            # Turn default checkers on.
            for checker in default_profile_checkers:
                self.set_checker_enabled(checker)

        self.enable_all = enable_all
        # If enable_all is given, almost all checkers should be enabled.
        disabled_groups = ["alpha.", "debug.", "osx.", "abseil-", "android-",
                           "darwin-", "objc-", "cppcoreguidelines-",
                           "fuchsia.", "fuchsia-", "hicpp-", "llvm-",
                           "llvmlibc-", "google-", "zircon-"]
        if enable_all:
            for checker_name, _ in checkers:
                if not any(checker_name.startswith(d_grp) for d_grp in
                           disabled_groups):

                    # There are a few exceptions, though, which still need to
                    # be manually enabled by the user: alpha and debug.
                    self.set_checker_enabled(checker_name)

                if checker_name.startswith("osx.") and \
                        platform.system() == 'Darwin':
                    # OSX checkers are only enable-all'd if we are on OSX.
                    self.set_checker_enabled(checker_name)

        # Set user defined enabled or disabled checkers from the command line.

        # Construct a list of reserved checker names.
        # (It is used to check if a profile name is valid.)
        reserved_names = self.__gen_name_variations()
        profiles = checker_labels.get_description('profile')
        guidelines = checker_labels.occurring_values('guideline')

        for identifier, enabled in cmdline_enable:
            if ':' in identifier:
                for checker in checker_labels.checkers_by_labels([identifier]):
                    self.set_checker_enabled(checker, enabled)
            elif identifier in profiles:
                if identifier in reserved_names:
                    LOG.warning("Profile name '%s' conflicts with a "
                                "checker(-group) name.", identifier)
                for checker in checker_labels.checkers_by_labels(
                        [f'profile:{identifier}']):
                    self.set_checker_enabled(checker, enabled)
            elif identifier in guidelines:
                if identifier in reserved_names:
                    LOG.warning("Guideline name '%s' conflicts with a "
                                "checker(-group) name.", identifier)
                for checker in checker_labels.checkers_by_labels(
                        [f'guideline:{identifier}']):
                    self.set_checker_enabled(checker, enabled)
            else:
                self.set_checker_enabled(identifier, enabled)



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
"""
Handle Thrift requests for configuration.
"""


import codechecker_api_shared

from codechecker_common.logger import get_logger

from codechecker_server.profiler import timeit

from codechecker_web.shared import convert

from ..database.config_db_model import Configuration
from ..database.database import DBSession

LOG = get_logger('server')


class ThriftConfigHandler:
    """
    Manages Thrift requests regarding configuration.
    """

    def __init__(self, auth_session, config_session):
        self.__auth_session = auth_session
        self.__session = config_session

    def __require_supermission(self):
        """
        Checks if the current user isn't a SUPERUSER.
        """
        if (not (self.__auth_session is None) and
                not self.__auth_session.is_root):
            raise codechecker_api_shared.ttypes.RequestFailed(
                codechecker_api_shared.ttypes.ErrorCode.UNAUTHORIZED,
                "You are not authorized to modify the notification.")

        return True

    @timeit
    def getNotificationBannerText(self):
        """
        Retrieves the notification banner text.
        """

        notificationString = ''
        with DBSession(self.__session) as session:
            notificationQuery = session.query(Configuration) \
                .filter(
                    Configuration.config_key == 'notification_banner_text') \
                .one_or_none()

            if notificationQuery is not None:
                notificationString = notificationQuery.config_value

        return convert.to_b64(notificationString)

    @timeit
    def setNotificationBannerText(self, notification_b64):
        """
        Sets the notification banner remove_products_except.
        Bevare: This method only works if the use is a SUPERUSER.
        """

        self.__require_supermission()

        notification = convert.from_b64(notification_b64)

        with DBSession(self.__session) as session:
            notificationQuery = session.query(Configuration) \
                .filter(
                    Configuration.config_key == 'notification_banner_text') \
                .one_or_none()

            if notificationQuery is None:
                conf = Configuration('notification_banner_text', notification)
                session.add(conf)
                session.flush()
            else:
                # update it
                notificationQuery.config_value = notification

            session.commit()
            session.close()



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------


import fnmatch
import os
from typing import List, Optional
import yaml

from codechecker_report_converter.report import Report, SourceReviewStatus
from codechecker_common.logger import get_logger
from codechecker_common.source_code_comment_handler import \
    SourceCodeCommentHandler, SpellException, contains_codechecker_comment, \
    SourceCodeComment, SourceCodeComments
from codechecker_common.util import path_for_fake_root


LOG = get_logger('system')


class ReviewStatusHandler:
    """
    This class helps to determine the review status of a report. The review
    status may come either from several sources:

    1. Source code comment:
        // codechecker_suppress [core.DivideZero] This is a false positive.
    2. Review status configuration file:
        <report_folder>/review_status.yaml contains review status settings on
        file path or bug has granularity.
    3. Review status rule:
        A mapping between bug hashes and review statuses set in the GUI.

    TODO: Option 3. should also be covered by this handler class.
    """

    REVIEW_STATUS_OPTIONS = [
        'false_positive',
        'suppress',
        'confirmed',
        'intentional']

    ALLOWED_FILTERS = [
        'filepath',
        'checker_name',
        'report_hash']

    # TODO: Read the commit message where THIS current line was committed!
    # That will tell you why "ignore" option is not found in this list.
    ALLOWED_ACTIONS = [
        'review_status',
        'reason']

    def __init__(self, source_root=""):
        """
        TODO: What happens if multiple report directories are stored?
        """
        self.__review_status_yaml = None
        self.__source_root = source_root
        self.__source_comment_warnings = []
        self.__source_commets = {}
        self.__data = None

    def __parse_codechecker_review_comment(
        self,
        source_file_name: str,
        report_line: int,
        checker_name: str
    ) -> SourceCodeComments:
        """Parse the CodeChecker review comments from a source file at a given
        position.  Returns an empty list if there are no comments.
        """
        src_comment_data = []
        with open(source_file_name, encoding='utf-8', errors='ignore') as f:
            if contains_codechecker_comment(f):
                sc_handler = SourceCodeCommentHandler()
                try:
                    src_comment_data = sc_handler.filter_source_line_comments(
                        f, report_line, checker_name)
                except SpellException as ex:
                    self.__source_comment_warnings.append(
                        f"{source_file_name} contains {ex}")

        return src_comment_data

    def __check_format_version_1(self):
        if 'rules' not in self.__data or \
                not isinstance(self.__data['rules'], list):
            raise ValueError(
                f"{self.__review_status_yaml} should contain the key 'rules' "
                f"with a non-empty list of (filters, actions) pair.")

        for rule in self.__data['rules']:
            if not isinstance(rule, dict) or \
                    'filters' not in rule or \
                    'actions' not in rule:
                raise ValueError(
                    f"Format error in {self.__review_status_yaml}. The "
                    f"following section should be an object with keys "
                    f"'filters' and 'actions':\n"
                    f"{yaml.dump(rule)}")

            if rule['filters'] is None:
                raise ValueError(
                    f"Format error in {self.__review_status_yaml}: 'filters' "
                    f"must have at least one element.")

            if rule['actions'] is None:
                raise ValueError(
                    f"Format error in {self.__review_status_yaml}: 'actions' "
                    f"must have at least one element.")

            for field in rule['filters']:
                if field not in ReviewStatusHandler.ALLOWED_FILTERS:
                    raise ValueError(
                        f"Format error in {self.__review_status_yaml}: filter "
                        f"'{field}' is not allowed. Available filters are: "
                        f"{', '.join(ReviewStatusHandler.ALLOWED_FILTERS)}")

            for field in rule['actions']:
                if field not in ReviewStatusHandler.ALLOWED_ACTIONS:
                    raise ValueError(
                        f"Format error in {self.__review_status_yaml}: action "
                        f"'{field}' is not allowed. Available actions are: "
                        f"{', '.join(ReviewStatusHandler.ALLOWED_ACTIONS)}")

            if 'review_status' not in rule['actions'] and \
                    'ignore' not in rule['actions']:
                raise ValueError(
                    f"Format error in {self.__review_status_yaml}. "
                    f"'review_status' is required in:\n"
                    f"{yaml.dump(rule)}.")

            if 'review_status' in rule['actions']:
                review_status_options = \
                    ', '.join(ReviewStatusHandler.REVIEW_STATUS_OPTIONS)

                if rule['actions']['review_status'] \
                        not in ReviewStatusHandler.REVIEW_STATUS_OPTIONS:
                    raise ValueError(
                        f"Invalid review status field: "
                        f"{rule['actions']['review_status']} at {rule} in "
                        f"{self.__review_status_yaml}. Available options are: "
                        f"{review_status_options}.")

                if rule['actions']['review_status'] == 'suppress':
                    rule['actions']['review_status'] = 'false_positive'

    def __validate_review_status_yaml_data(self):
        """
        This function validates the data read from review_status.yaml file and
        raises ValueError with the description of the error if the format is
        invalid.
        """
        if not isinstance(self.__data, dict):
            raise ValueError(
                f"{self.__review_status_yaml} should represent a dictionary.")

        if '$version' not in self.__data:
            raise ValueError(
                f"{self.__review_status_yaml} must contain the key "
                f"'$version'. Hint: Currently the latest version of this "
                f"is 1. Consider adding '$version: 1' to the top of the file.")

        if not isinstance(self.__data['$version'], int):
            raise ValueError(
                f"{self.__review_status_yaml} should have an integer value "
                "for the key '$version'.")

        if self.__data['$version'] == 1:
            self.__check_format_version_1()

    def __report_matches_rule(self, report: Report, rule: dict):
        if 'filepath' in rule['filters'] and not fnmatch.fnmatch(
                report.file.original_path, rule['filters']['filepath']):
            return False

        if 'checker_name' in rule['filters'] and \
                report.checker_name != rule['filters']['checker_name']:
            return False

        if 'report_hash' in rule['filters'] and not \
                report.report_hash.startswith(rule['filters']['report_hash']):
            return False

        return True

    def get_review_status(self, report: Report) -> SourceReviewStatus:
        """
        Return the review status of the report based on source code comments.

        If the review status is not set in the source code then "unreviewed"
        review status returns. This function throws ValueError if the review
        status is ambiguous (see get_review_status_from_source()).
        """
        # 1. Check if the report has in-source review status setting.
        review_status = self.get_review_status_from_source(report)

        if review_status:
            return review_status

        # 2. Check if the report has review status setting in the yaml config.
        if self.__data:
            review_status = self.get_review_status_from_config(report)
            if review_status:
                return review_status

        # 3. Return "unreviewed" otherwise.
        return SourceReviewStatus(bug_hash=report.report_hash)

    def set_review_status_config(self, config_file):
        """
        Set the location of the review_status.yaml config file.

        When the content of multiple report directories are parsed then they
        may contain separate config files. These need to be given before
        parsing each report folders.
        """
        self.__review_status_yaml = config_file

        if os.path.islink(self.__review_status_yaml):
            orig_yaml = os.readlink(self.__review_status_yaml)
            if not os.path.exists(orig_yaml):
                raise ValueError(f"{orig_yaml} not found.")

        with open(self.__review_status_yaml,
                  encoding='utf-8', errors='ignore') as f:
            # TODO: Validate format.
            #  - Can filepath be a list?
            try:
                self.__data = yaml.safe_load(f)
            except yaml.YAMLError as err:
                raise ValueError(
                    f"Invalid YAML format in {self.__review_status_yaml}:\n"
                    f"{err}")

        self.__validate_review_status_yaml_data()

    def should_ignore(self, report: Report) -> bool:
        """
        This function returns True if the Report should be ignored based on the
        review status config file. If a report is ignored, then it shouldn't be
        written to the analysis output files and CodeChecker should consider
        them as not existing reports in any context.

        TODO: If "ignore" is a standalone field, then maybe we should find
            another name for this class, because it handles not only review
            statuses.
        """
        if self.__data is None:
            return False

        for rule in self.__data['rules']:
            if self.__report_matches_rule(report, rule) and \
                    rule['actions'].get('ignore'):
                return True

        return False

    def get_review_status_from_config(
        self,
        report: Report
    ) -> Optional[SourceReviewStatus]:
        """
        Return the review status of the given report based on the config file
        set by set_review_status_config(). If not config file set, or no
        setting matches the report then None returns.
        """
        assert self.__data is not None, \
            "Review status config file has to be set with " \
            "set_review_status_config()."

        # TODO: Document "in_source".
        for rule in self.__data['rules']:
            if not self.__report_matches_rule(report, rule) or \
                    rule['actions'].get('ignore'):
                continue

            if any(filt in rule['filters'] for filt in
                   ['filepath', 'checker_name', 'report_hash']):
                return SourceReviewStatus(
                    status=rule['actions']['review_status'],
                    message=rule['actions']['reason']
                    .encode(encoding='utf-8', errors='ignore')
                    if 'reason' in rule['actions'] else b'',
                    bug_hash=report.report_hash,
                    in_source=True)

        return None

    def get_review_status_from_source(
        self,
        report: Report
    ) -> Optional[SourceReviewStatus]:
        """
        Return the review status based on the source code comment belonging to
        the given report.

        - Return the review status if it is set in the source code.
        - If the review status is ambiguous (i.e. multiple source code
            comments belong to it) then a ValueError exception is raised which
            contains information about the problem in a string.
        - If the soure file changed (which means that the source comments may
            have replaced, changed or removed) then None returns.
        - If no review status belongs to the report in the source code, then
            return None.
        TODO: This function either returns a SourceReviewStatus, or raises an
        exception or returns None. This is too many things that a caller needs
        to handle. The reason is that according to the current behaviors,
        ambiguous comment results a hard error, and changed files result a
        warning with "unreviewed" status. In the future we could implement a
        logic where we take the first comment into account in case of ambiguity
        or return "unreviewed" with a warning, like changed files.
        """
        if self.__source_root:
            # sanitize paths like mass store run
            source_file_name = path_for_fake_root(
                report.file.original_path, self.__source_root)
        else:
            # in this branch we are operating on the original filesystem,
            # and we need the path to the original file
            source_file_name = os.path.realpath(os.path.join(
                self.__source_root, report.file.original_path))

        if source_file_name in report.changed_files:
            return None

        if os.path.isfile(source_file_name):
            src_comment_data = self.__parse_codechecker_review_comment(
                source_file_name, report.line, report.checker_name)

            if len(src_comment_data) == 1:
                data = src_comment_data[0]
                status, message = data.status, data.message
                self.__source_commets[report] = data

                return SourceReviewStatus(
                    status=status,
                    message=message.encode('utf-8'),
                    bug_hash=report.report_hash,
                    in_source=True)
            elif len(src_comment_data) > 1:
                raise ValueError(
                    f"Multiple source code comments can be found for "
                    f"'{report.checker_name}' checker in '{source_file_name}' "
                    f"at line {report.line}.")

        return None

    def source_comment_warnings(self) -> List[str]:
        """
        Sometimes it is not intuitive why the given review status is determined
        for a report. For example, if an in-source suppression is misspelled,
        then the report remains unreviewed:
        // codechecker_suppresssss [<checker name>] comment
        This function returns a list of warning messages on these unintuitive
        behaviors.
        """
        return self.__source_comment_warnings

    def source_comment(self, report: Report) -> Optional[SourceCodeComment]:
        """
        This ReviewStatusHandler class is caching source comments so they are
        read and parsed only once for each report.
        """
        return self.__source_commets.get(report)



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
"""
Arranges the 1st phase of 2 phase executions for CTU
"""


import glob
import yaml
import os
import shutil
import tempfile
from pathlib import Path
from sys import maxsize
from yaml import Dumper

from codechecker_common.logger import get_logger

from codechecker_merge_clang_extdef_mappings.merge_clang_extdef_mappings \
    import merge

from .. import analyzer_base
from . import ctu_triple_arch

LOG = get_logger('analyzer')


class LLVMComatibleYamlDumper(Dumper):
    def check_simple_key(self):
        """ Mark every keys as simple keys.

        PyYAML limits simple keys to '128' characters and this value can't be
        changed (https://github.com/yaml/pyyaml/issues/157). To be compatible
        with the YAML parser of LLVM we override this function and mark every
        keys as simple keys.
        """
        return True


def merge_clang_extdef_mappings(ctu_dir, ctu_func_map_file,
                                ctu_temp_fnmap_folder):
    """ Merge individual function maps into a global one."""

    triple_arches = glob.glob(os.path.join(ctu_dir, '*'))
    for triple_path in triple_arches:
        if not os.path.isdir(triple_path):
            continue

        triple_arch = os.path.basename(triple_path)
        fnmap_dir = os.path.join(ctu_dir, triple_arch,
                                 ctu_temp_fnmap_folder)

        merged_fn_map = os.path.join(ctu_dir, triple_arch,
                                     ctu_func_map_file)
        merge(fnmap_dir, merged_fn_map)

        # Remove all temporary files.
        shutil.rmtree(fnmap_dir, ignore_errors=True)


def generate_ast_cmd(action, config, triple_arch, source):
    """ Command to generate AST (or PCH) file. """
    ast_joined_path = os.path.join(config.ctu_dir, triple_arch, 'ast',
                                   os.path.realpath(source)[1:] + '.ast')
    ast_path = os.path.abspath(ast_joined_path)
    ast_dir = os.path.dirname(ast_path)

    cmd = ctu_triple_arch.get_compile_command(action, config, source)

    # __clang__analyzer__ macro needs to be set in the imported TUs too.
    cmd.extend(['-emit-ast', '-D__clang_analyzer__', '-w', '-o', ast_path])

    return cmd, ast_dir


def generate_invocation_list(triple_arch, action, source, config):
    """ Generates the invocation for the source file of the current
    compilation command. Used during on-demand analysis. The invocation list
    is a mapping from absolute paths of the source files to the parts of the
    compilation commands that are used to generate from each one of them. """

    triple_arch_dir = Path(config.ctu_dir, triple_arch)
    # Ensure, that architecture directory exists.
    triple_arch_dir.mkdir(parents=True, exist_ok=True)

    invocation_list = triple_arch_dir / 'invocation-list.yml'

    source_path = Path(source).resolve()

    cmd = ctu_triple_arch.get_compile_command(action, config, source)
    # __clang__analyzer__ macro needs to be set in the imported TUs too.
    cmd.extend(['-D__clang_analyzer__', '-w'])

    # The YAML mapping entry already has a newline at the end.
    # Line width is set to max int size because of compatibility with the YAML
    # parser of LLVM. We try to ensure that no lines break in the textual
    # representation of the list items.
    invocation_line = yaml.dump(
        {str(source_path): cmd},
        width=maxsize, Dumper=LLVMComatibleYamlDumper)

    LOG.debug_analyzer("Appending invocation list item '%s'", invocation_line)

    # Append the next entry into the invocation list yaml.
    with invocation_list.open('a', encoding='utf-8', errors='ignore') as \
            invocation_file:
        invocation_file.write(invocation_line)


def generate_ast(triple_arch, action, source, config):
    """ Generates ASTs for the current compilation command. Used during
    ast-dump based analysis. """

    cmd, ast_dir = generate_ast_cmd(action, config, triple_arch, source)

    if not os.path.isdir(ast_dir):
        try:
            os.makedirs(ast_dir)
        except OSError:
            pass

    cmdstr = ' '.join(cmd)
    LOG.debug_analyzer("Generating AST using '%s'", cmdstr)
    ret_code, _, err = \
        analyzer_base.SourceAnalyzer.run_proc(cmd, action.directory)

    if ret_code != 0:
        LOG.error("Error generating AST.\n\ncommand:\n\n%s\n\nstderr:\n\n%s",
                  cmdstr, err)


def ast_dump_path(source_path):
    """ AST-dump based analysis uses preprocessed paths, here the path prefix
    'ast' and the filename suffix '.ast' is added to the name of the original
    source file. """

    # Normalize path on Windows OS.
    path = os.path.splitdrive(source_path)[1]
    # Make relative path out of absolute.
    path = path[1:] if path[0] == os.sep else path
    # Prepend path segment, and append filename suffix.
    return os.path.join("ast", path + ".ast")


def func_map_list_src_to_ast(func_src_list, ctu_on_demand):
    """ Turns textual function map list with source files into a
    mapping from mangled names to mapped paths, which can be absolute paths to
    the original source files if ctu_on_demand is True, or relative path
    segments to AST-dump files that reside in CTU-DIR directory otherwise. """

    func_ast_list = []
    for fn_src_txt in func_src_list:
        # FIXME: Detect and report invalid input.
        # The format of the external function map file changed in between
        # clang-15 and clang-16, check whether this is the updated format.
        if fn_src_txt[0].isdigit():
            length_str, _ = fn_src_txt.split(':', 1)
            length = int(length_str)
            sep_pos = len(length_str) + 1 + length
            mangled_name = fn_src_txt[0: sep_pos]
            path = fn_src_txt[sep_pos + 1:]  # Skipping the ' ' separator
        else:
            dpos = fn_src_txt.find(" ")
            mangled_name = fn_src_txt[0:dpos]
            path = fn_src_txt[dpos + 1:]

        # On-demand analysis does not require any preprocessing on the source
        # file paths, contrary to AST-dump based.
        mapped_path = path if ctu_on_demand else ast_dump_path(path)

        func_ast_list.append(mangled_name + " " + mapped_path)
    return func_ast_list


def get_extdef_mapping_cmd(action, config, source, func_map_cmd):
    """ Get command to create CTU index file. """

    cmd = ctu_triple_arch.get_compile_command(action, config)
    cmd[0] = func_map_cmd
    cmd.insert(1, source)
    cmd.insert(2, '--')
    return cmd


def map_functions(triple_arch, action, source, config,
                  func_map_cmd, temp_fnmap_folder):
    """ Generate function map file for the current source.

        On-demand CTU analysis requires the *mangled name* to *source file*
        mapping. However in case of pre-processed ast-dumps, *mangled name* to
        *ast dump* mapping must be provided.
    """

    cmd = get_extdef_mapping_cmd(action, config, source, func_map_cmd)

    cmdstr = ' '.join(cmd)
    LOG.debug_analyzer("Generating function map using '%s'", cmdstr)
    ret_code, stdout, err \
        = analyzer_base.SourceAnalyzer.run_proc(cmd, action.directory)
    if ret_code != 0:
        LOG.error("Error generating function map."
                  "\n\ncommand:\n\n%s\n\nstderr:\n\n%s", cmdstr, err)
        return

    func_src_list = stdout.splitlines()
    func_ast_list = func_map_list_src_to_ast(
        func_src_list, config.ctu_on_demand)
    extern_fns_map_folder = os.path.join(config.ctu_dir, triple_arch,
                                         temp_fnmap_folder)
    if not os.path.isdir(extern_fns_map_folder):
        try:
            os.makedirs(extern_fns_map_folder)
        except OSError:
            pass

    if func_ast_list:
        with tempfile.NamedTemporaryFile(mode='w',
                                         dir=extern_fns_map_folder,
                                         delete=False,
                                         encoding='utf-8') as out_file:
            out_file.write("\n".join(func_ast_list) + "\n")



"""
Remove db_version table

Revision ID: a79677f54e48
Revises:     6cb6a3a41967
Create Date: 2020-03-13 12:14:19.805990
"""

from alembic import op
import sqlalchemy as sa


# Revision identifiers, used by Alembic.
revision = 'a79677f54e48'
down_revision = '6cb6a3a41967'
branch_labels = None
depends_on = None


def upgrade():
    op.drop_table('db_version')


def downgrade():
    op.create_table('db_version',
                    sa.Column('major', sa.INTEGER(), nullable=False),
                    sa.Column('minor', sa.INTEGER(), nullable=False),
                    sa.PrimaryKeyConstraint('major', 'minor',
                                            name='pk_db_version'))



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
"""
Config handler for Clang Tidy analyzer.
"""


from codechecker_common.logger import get_logger

from ..config_handler import AnalyzerConfigHandler, CheckerState, \
                             get_compiler_warning_name_and_type


def is_compiler_warning(checker_name):
    name, _ = get_compiler_warning_name_and_type(checker_name)
    return name is not None


LOG = get_logger('analyzer.tidy')


class ClangTidyConfigHandler(AnalyzerConfigHandler):
    """
    Configuration handler for Clang-tidy analyzer.
    """

    def __init__(self):
        super(ClangTidyConfigHandler, self).__init__()

    def add_checker(self, checker_name, description='',
                    state=CheckerState.disabled):
        """
        Add additional checker if the 'take-config-from-directory'
        analyzer configuration option is not set.
        """
        if self.analyzer_config and \
           self.analyzer_config.get('take-config-from-directory') == 'true':
            if is_compiler_warning(checker_name):
                return

        super(ClangTidyConfigHandler, self).add_checker(checker_name,
                                                        description, state)

    def set_checker_enabled(self, checker_name, enabled=True):
        """
        Enable checker, keep description if already set.
        """
        if checker_name.startswith('W') or \
           checker_name.startswith('clang-diagnostic'):
            self.add_checker(checker_name)

        super(ClangTidyConfigHandler, self).set_checker_enabled(checker_name,
                                                                enabled)



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
""""""


import os
import re

from codechecker_analyzer import analyzer_context
from codechecker_common.logger import get_logger

LOG = get_logger('system')


def get_log_env(logfile, original_env):
    """
    Environment for logging. With the ld logger.
    Keep the original environment unmodified as possible.
    Only environment variables required for logging are changed.
    """
    context = analyzer_context.get_context()
    new_env = original_env

    new_env[context.env_var_cc_logger_bin] = context.path_logger_bin

    new_env['LD_PRELOAD'] = context.logger_lib_path

    # Set ld logger logfile.
    new_env[context.env_var_cc_logger_file] = logfile

    return new_env


def extend(path_env_extra, ld_lib_path_extra):
    """Extend the checker environment.

    The default environment is extended with the given PATH and
    LD_LIBRARY_PATH values to find tools if they ar not on
    the default places.
    """
    new_env = os.environ.copy()

    if path_env_extra:
        extra_path = ':'.join(path_env_extra)
        LOG.debug_analyzer(
            'Extending PATH environment variable with: ' + extra_path)

        try:
            new_env['PATH'] = extra_path + ':' + new_env['PATH']
        except KeyError:
            new_env['PATH'] = extra_path

    if ld_lib_path_extra:
        extra_lib = ':'.join(ld_lib_path_extra)
        LOG.debug_analyzer(
            'Extending LD_LIBRARY_PATH environment variable with: ' +
            extra_lib)
        try:
            original_ld_library_path = new_env['LD_LIBRARY_PATH']
            new_env['LD_LIBRARY_PATH'] = \
                extra_lib + ':' + original_ld_library_path
        except KeyError:
            new_env['LD_LIBRARY_PATH'] = extra_lib

    return new_env


def replace_env_var(cfg_file):
    """
    Returns a replacement function which can be used in RegEx functions such as
    re.sub to replace matches with a string from the OS environment.
    """
    def replacer(matchobj):
        env_var = matchobj.group(1)
        if env_var not in os.environ:
            LOG.error('%s environment variable not set in %s', env_var,
                      cfg_file)
            return ''
        return os.environ[env_var]

    return replacer


def find_by_regex_in_envpath(pattern, environment):
    """
    Searches for files matching the pattern string in the environment's PATH.
    """

    regex = re.compile(pattern)

    binaries = {}
    for path in environment['PATH'].split(os.pathsep):
        _, _, filenames = next(os.walk(path), ([], [], []))
        for f in filenames:
            if re.match(regex, f):
                if binaries.get(f) is None:
                    binaries[f] = [os.path.join(path, f)]
                else:
                    binaries[f].append(os.path.join(path, f))

    return binaries


def get_binary_in_path(basename_list, versioning_pattern, env):
    """
    Select the most matching binary for the given pattern in the given
    environment. Works well for binaries that contain versioning.
    """

    binaries = find_by_regex_in_envpath(versioning_pattern, env)

    if not binaries:
        return False
    elif len(binaries) == 1:
        # Return the first found (earliest in PATH) binary for the only
        # found binary name group.
        return list(binaries.values())[0][0]
    else:
        keys = list(binaries.keys())
        keys.sort()

        # If one of the base names match, select that version.
        files = None
        for base_key in basename_list:
            # Cannot use set here as it would destroy precendence.
            if base_key in keys:
                files = binaries[base_key]
                break

        if not files:
            # Select the "newest" available version if there are multiple and
            # none of the base names matched.
            files = binaries[keys[-1]]

        # Return the one earliest in PATH.
        return files[0]


def is_analyzer_from_path():
    """ Return True if CC_ANALYZERS_FROM_PATH environment variable is set. """
    analyzers_from_path = os.environ.get('CC_ANALYZERS_FROM_PATH', '').lower()
    if analyzers_from_path in ['yes', '1']:
        return True
    return False


def get_clangsa_plugin_dir():
    """ Return the value of the CC_CLANGSA_PLUGIN_DIR environment variable. """
    return os.environ.get('CC_CLANGSA_PLUGIN_DIR')



"""
Share the locking of runs across servers via database

Revision ID: dd9c97ead24
Revises:     4b38fa14c27b
Create Date: 2017-11-17 15:44:07.810579
"""

from alembic import op
import sqlalchemy as sa


# Revision identifiers, used by Alembic.
revision = 'dd9c97ead24'
down_revision = '4b38fa14c27b'
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        'run_locks',
        sa.Column('name', sa.String(), nullable=False),
        sa.Column('locked_at', sa.DateTime(), nullable=False),
        sa.Column('username', sa.String(), nullable=True),
        sa.PrimaryKeyConstraint('name', name=op.f('pk_run_locks'))
    )


def downgrade():
    op.drop_table('run_locks')
